import enum
from typing import Optional, List, Generic, Union, Tuple

from causy.contrib.graph_ui import (
    EdgeTypeConfig,
    EdgeUIConfig,
    ConditionalEdgeUIConfig,
    ConditionalEdgeUIConfigComparison,
)
from causy.interfaces import (
    EdgeTypeInterface,
    EdgeTypeInterfaceType,
    NodeInterface,
)


class EdgeTypeEnum(str, enum.Enum):
    DIRECTED = "directed"
    UNDIRECTED = "undirected"
    BIDIRECTED = "bidirected"


class DirectedEdge(EdgeTypeInterface, Generic[EdgeTypeInterfaceType]):
    name: str = EdgeTypeEnum.DIRECTED.name
    IS_DIRECTED: bool = True
    STR_REPRESENTATION: str = "-->"  # u --> v

    class GraphAccessMixin:
        def only_directed_edge_exists(
            self, u: Union[NodeInterface, str], v: Union[NodeInterface, str]
        ) -> bool:
            """
            Check if a directed edge exists between u and v, but no directed edge exists between v and u. Case: u -> v
            :param u: node u
            :param v: node v
            :return: True if only directed edge exists, False otherwise
            """
            if self.directed_edge_exists(u, v) and not self.directed_edge_exists(v, u):
                return True
            return False

        def _resolve_node_references(
            self,
            u: Union[NodeInterface, str],
            v: Optional[Union[NodeInterface, str]] = None,
        ) -> Union[NodeInterface, Tuple[NodeInterface, NodeInterface]]:
            """
            Resolve node references
            :param u:
            :param v:
            :return: Returns a tuple of nodes if v is not None, otherwise returns a single node
            """
            if isinstance(u, str):
                u = self.node_by_id(u)
            if v and isinstance(v, str):
                v = self.node_by_id(v)

            if v is None:
                return u

            return u, v

        def directed_paths(
            self, u: Union[NodeInterface, str], v: Union[NodeInterface, str]
        ) -> List[List[Tuple[NodeInterface, NodeInterface]]]:
            """
            Return all directed paths from u to v
            :param u: node u
            :param v: node v
            :return: list of directed paths
            """
            u, v = self._resolve_node_references(u, v)
            # TODO: try a better data structure for this
            if self.directed_edge_exists(u, v):
                return [[(u, v)]]
            paths = []
            for w in self.edges[u.id]:
                if self.directed_edge_exists(u, self.nodes[w]):
                    for path in self.directed_paths(self.nodes[w], v):
                        paths.append([(u, self.nodes[w])] + path)
            return paths


class DirectedEdgeUIConfig(EdgeTypeConfig):
    edge_type: str = DirectedEdge().name

    default_ui_config: Optional[EdgeUIConfig] = EdgeUIConfig(
        label_field="correlation",
        color="#333",
        width=4,
        style="dashed",
        animated=True,
        marker_start=None,
        marker_end="ArrowClosed",
        label="Direct Effect: ${correlation.toFixed(4)}",
    )
    conditional_ui_configs: Optional[List[ConditionalEdgeUIConfig]] = [
        ConditionalEdgeUIConfig(
            ui_config=EdgeUIConfig(
                color="#f00000",
                width=4,
                style="dashed",
                animated=True,
                marker_start=None,
                marker_end="ArrowClosed",
                label_field="direct_effect",
                label="Direct Effect: ${direct_effect.toFixed(4)}",
            ),
            condition_field="direct_effect",
            condition_value=0,
            condition_comparison=ConditionalEdgeUIConfigComparison.LESS,
        ),
        ConditionalEdgeUIConfig(
            ui_config=EdgeUIConfig(
                color="#0f0fff",
                width=4,
                style="dashed",
                animated=True,
                marker_start=None,
                marker_end="ArrowClosed",
                label_field="direct_effect",
                label="Direct Effect: ${direct_effect.toFixed(4)}",
            ),
            condition_field="direct_effect",
            condition_value=0,
            condition_comparison=ConditionalEdgeUIConfigComparison.GREATER,
        ),
    ]


class UndirectedEdge(EdgeTypeInterface, Generic[EdgeTypeInterfaceType]):
    name: str = EdgeTypeEnum.UNDIRECTED.name
    IS_DIRECTED: bool = False
    STR_REPRESENTATION: str = "---"  # u --- v


class UndirectedEdgeUIConfig(EdgeTypeConfig):
    edge_type: str = UndirectedEdge().name

    default_ui_config: Optional[EdgeUIConfig] = EdgeUIConfig(
        label_field="direct_effect",
        color="#333",
        width=4,
        style="solid",
        animated=False,
        marker_start=None,
        marker_end=None,
        label="Correlation: ${correlation.toFixed(4)}",
    )


class BiDirectedEdge(EdgeTypeInterface, Generic[EdgeTypeInterfaceType]):
    name: str = EdgeTypeEnum.BIDIRECTED.name
    IS_DIRECTED: bool = False  # This is a bi-directed edge - so it is not directed in the traditional sense
    STR_REPRESENTATION: str = "<->"  # u <-> v


class BiDirectedEdgeUIConfig(EdgeTypeConfig):
    edge_type: str = BiDirectedEdge().name

    default_ui_config: Optional[EdgeUIConfig] = EdgeUIConfig(
        label_field="direct_effect",
        color="#333",
        width=4,
        style="solid",
        animated=False,
        marker_start="ArrowClosed",
        marker_end="ArrowClosed",
    )


EDGE_TYPES = {
    DirectedEdge().name: DirectedEdge,
    UndirectedEdge().name: UndirectedEdge,
    BiDirectedEdge().name: BiDirectedEdge,
}



import multiprocessing
from abc import ABC, abstractmethod
from typing import List, Dict, Optional, Union, TypeVar, Generic, Any
from typing_extensions import Annotated
import logging

from pydantic import BaseModel, computed_field, Field, PlainValidator, WithJsonSchema
import torch

from causy.graph_utils import (
    load_pipeline_artefact_by_definition,
    serialize_module_name,
    load_pipeline_steps_by_definition,
)
from causy.variables import (
    StringParameter,
    IntegerParameter,
    BoolParameter,
    FloatParameter,
)

logger = logging.getLogger(__name__)

DEFAULT_THRESHOLD = 0.01

AS_MANY_AS_FIELDS = 0

MetadataBaseType = Union[str, int, float, bool]
MetadataType = Union[
    str, int, float, bool, List[MetadataBaseType], Dict[str, MetadataBaseType]
]


class ComparisonSettingsInterface(BaseModel, ABC):
    min: IntegerParameter
    max: IntegerParameter

    @computed_field
    @property
    def name(self) -> str:
        return serialize_module_name(self)


class NodeInterface(BaseModel, ABC):
    """
    Node interface for the graph. A node is defined by a name and a value.
    """

    name: str
    id: str
    values: Annotated[
        Optional[torch.DoubleTensor],
        WithJsonSchema(
            {
                "type": "array",
                "items": {"type": "number"},
                "description": "Node values",
            }
        ),
    ] = Field(exclude=True, default=None)

    class Config:
        arbitrary_types_allowed = True


EdgeTypeInterfaceType = TypeVar("EdgeTypeInterfaceType")


class EdgeTypeInterface(ABC, BaseModel, Generic[EdgeTypeInterfaceType]):
    """
    Edge type interface for the graph
    An edge type is defined by a name
    """

    name: str

    # define if it is a directed or undirected edge type (default is undirected). We use this e.g. when we compare the graph.
    IS_DIRECTED: bool = True
    STR_REPRESENTATION: str = "-"

    class GraphAccessMixin(ABC):
        pass

    def pre_edge_remove_hook(self, graph: "BaseGraphInterface", edge: "EdgeInterface"):
        """
        Hook that is executed before an edge is deleted.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def post_edge_remove_hook(self, graph: "BaseGraphInterface", edge: "EdgeInterface"):
        """
        Hook that is executed after an edge is deleted.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def pre_edge_add_hook(self, graph: "BaseGraphInterface", edge: "EdgeInterface"):
        """
        Hook that is executed before an edge is added.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def post_edge_add_hook(self, graph: "BaseGraphInterface", edge: "EdgeInterface"):
        """
        Hook that is executed after an edge is added.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def pre_edge_update_hook(self, graph: "BaseGraphInterface", edge: "EdgeInterface"):
        """
        Hook that is executed before an edge is updated.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def post_edge_update_hook(self, graph: "BaseGraphInterface", edge: "EdgeInterface"):
        """
        Hook that is executed after an edge is updated.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def pre_edge_type_change_hook(
        self, graph: "BaseGraphInterface", edge: "EdgeInterface"
    ):
        """
        Hook that is executed before an edge type is changed.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def post_edge_type_change_hook(
        self, graph: "BaseGraphInterface", edge: "EdgeInterface"
    ):
        """
        Hook that is executed after an edge type is changed.
        :param graph: the graph
        :param edge: the edge
        """
        pass

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class EdgeInterface(BaseModel, ABC):
    """
    Edge interface for the graph
    A graph edge is defined by two nodes and an edge type. It can also have metadata.
    """

    u: NodeInterface
    v: NodeInterface
    edge_type: EdgeTypeInterface
    metadata: Dict[str, MetadataType] = None

    class Config:
        arbitrary_types_allowed = True

    def __eq__(self, other):
        """
        Check if two edges are equal by comparing the nodes and the edge type
        :param other:
        :return:
        """
        if not isinstance(other, EdgeInterface):
            return False

        if self.edge_type != other.edge_type:
            return False
        if self.edge_type.IS_DIRECTED:
            return self.u == other.u and self.v == other.v
        else:
            return self.is_connection_between_same_nodes(other)

    def is_connection_between_same_nodes(self, edge):
        return (
            self.u == edge.u
            and self.v == edge.v
            or self.u == edge.v
            and self.v == edge.u
        )


class TestResultInterface(BaseModel, ABC):
    """
    Test result interface for the graph
    A test result is defined by two nodes and an action. It can also have metadata.
    """

    u: NodeInterface
    v: NodeInterface
    action: str
    data: Optional[Dict] = None

    class Config:
        arbitrary_types_allowed = True


class BaseGraphInterface(ABC):
    nodes: Dict[str, NodeInterface]
    edges: Dict[str, Dict[str, Dict]]

    @abstractmethod
    def retrieve_edge_history(self, u, v, action: str) -> List[TestResultInterface]:
        pass

    @abstractmethod
    def add_edge_history(self, u, v, action: TestResultInterface):
        pass

    @abstractmethod
    def add_edge(self, u, v, metadata):
        pass

    @abstractmethod
    def remove_edge(self, u, v):
        pass

    @abstractmethod
    def remove_directed_edge(self, u, v):
        pass

    @abstractmethod
    def update_edge(self, u, v, metadata=None, edge_type=None):
        pass

    @abstractmethod
    def update_directed_edge(self, u, v, metadata=None, edge_type=None):
        pass

    @abstractmethod
    def add_node(self, name, values) -> NodeInterface:
        pass

    @abstractmethod
    def edge_value(self, u, v):
        pass

    @abstractmethod
    def undirected_edge_exists(self, u, v):
        pass

    @abstractmethod
    def directed_edge_exists(self, u, v):
        pass

    @abstractmethod
    def edge_exists(self, u, v, ignore_soft_deleted=False):
        pass

    @abstractmethod
    def restore_edge(self, u, v):
        pass

    @abstractmethod
    def restore_directed_edge(self, u, v):
        pass


class GraphModelInterface(ABC):
    pool: multiprocessing.Pool

    @abstractmethod
    def create_graph_from_data(self, data: List[Dict]):
        pass

    @abstractmethod
    def execute_pipeline_steps(self):
        pass

    @abstractmethod
    def execute_pipeline_step(self, step, apply_to_graph: bool = True):
        pass


class GeneratorInterface(ABC, BaseModel):
    comparison_settings: Optional[ComparisonSettingsInterface] = None
    chunked: Optional[BoolParameter] = False
    every_nth: Optional[IntegerParameter] = None
    generator: Optional["GeneratorInterface"] = None
    shuffle_combinations: Optional[BoolParameter] = None

    @abstractmethod
    def generate(self, graph: BaseGraphInterface, graph_model_instance_: dict):
        pass

    def __init__(
        self,
        comparison_settings: Optional[ComparisonSettingsInterface] = None,
        chunked: Optional[BoolParameter] = None,
        every_nth: Optional[IntegerParameter] = None,
        generator: Optional["GeneratorInterface"] = None,
        shuffle_combinations: Optional[BoolParameter] = None,
    ) -> None:
        super().__init__(comparison_settings=comparison_settings)
        if isinstance(comparison_settings, dict):
            comparison_settings = load_pipeline_artefact_by_definition(
                comparison_settings
            )

        if chunked is not None:
            self.chunked = chunked

        if every_nth is not None:
            self.every_nth = every_nth

        if generator is not None:
            self.generator = generator

        if shuffle_combinations is not None:
            self.shuffle_combinations = shuffle_combinations

        self.comparison_settings = comparison_settings

    @computed_field
    @property
    def name(self) -> str:
        return serialize_module_name(self)


PipelineStepInterfaceType = TypeVar("PipelineStepInterfaceType")


class PipelineStepInterface(ABC, BaseModel, Generic[PipelineStepInterfaceType]):
    generator: Optional[GeneratorInterface] = None
    threshold: Optional[FloatParameter] = DEFAULT_THRESHOLD
    chunk_size_parallel_processing: IntegerParameter = 1
    parallel: BoolParameter = True

    display_name: Optional[StringParameter] = None

    needs_unapplied_actions: Optional[BoolParameter] = False

    def __init__(
        self,
        threshold: Optional[FloatParameter] = None,
        generator: Optional[GeneratorInterface] = None,
        chunk_size_parallel_processing: Optional[IntegerParameter] = None,
        parallel: Optional[BoolParameter] = None,
        display_name: Optional[StringParameter] = None,
        **kwargs,
    ) -> None:
        super().__init__()
        if generator:
            if isinstance(generator, dict):
                self.generator = load_pipeline_artefact_by_definition(generator)
            else:
                self.generator = generator

        if chunk_size_parallel_processing:
            self.chunk_size_parallel_processing = chunk_size_parallel_processing

        if parallel:
            self.parallel = parallel

        if display_name:
            self.display_name = display_name

        if threshold:
            self.threshold = threshold

    @computed_field
    @property
    def name(self) -> str:
        return serialize_module_name(self)

    @abstractmethod
    def process(
        self,
        nodes: List[str],
        graph: BaseGraphInterface,
        unapplied_actions: Optional[List[TestResultInterface]] = None,
    ) -> Optional[TestResultInterface]:
        """
        Test if u and v are independent
        :param u: u values
        :param v: v values
        :return: True if independent, False otherwise
        """
        pass

    def __call__(
        self,
        nodes: List[str],
        graph: BaseGraphInterface,
        unapplied_actions: Optional[List[TestResultInterface]] = None,
    ) -> Optional[TestResultInterface]:
        if self.needs_unapplied_actions and unapplied_actions is None:
            logger.warn(
                f"Pipeline step {self.name} needs unapplied actions but none were provided"
            )
        elif self.needs_unapplied_actions and unapplied_actions is not None:
            return self.process(nodes, graph, unapplied_actions)

        return self.process(nodes, graph)


class ExitConditionInterface(ABC, BaseModel):
    @abstractmethod
    def check(
        self,
        graph: BaseGraphInterface,
        graph_model_instance_: GraphModelInterface,
        actions_taken: List[TestResultInterface],
        iteration: IntegerParameter,
    ) -> bool:
        """
        :param graph:
        :param graph_model_instance_:
        :param actions_taken:
        :param iteration:
        :return: True if you want to break an iteration, False otherwise
        """
        pass

    def __call__(
        self,
        graph: BaseGraphInterface,
        graph_model_instance_: GraphModelInterface,
        actions_taken: List[TestResultInterface],
        iteration: int,
    ) -> bool:
        return self.check(graph, graph_model_instance_, actions_taken, iteration)

    @computed_field
    @property
    def name(self) -> str:
        return serialize_module_name(self)


LogicStepInterfaceType = TypeVar("LogicStepInterfaceType")


class LogicStepInterface(ABC, BaseModel, Generic[LogicStepInterfaceType]):
    pipeline_steps: Optional[List[Union[PipelineStepInterfaceType]]] = None
    exit_condition: Optional[ExitConditionInterface] = None

    display_name: Optional[StringParameter] = None

    @abstractmethod
    def execute(self, graph: BaseGraphInterface, graph_model_instance_: dict):
        pass

    @computed_field
    @property
    def name(self) -> str:
        return serialize_module_name(self)

    def __init__(
        self,
        pipeline_steps: Optional[
            Union[List[PipelineStepInterfaceType], Dict[Any, Any]]
        ] = None,
        exit_condition: Union[ExitConditionInterface, Dict[Any, Any]] = None,
        display_name: Optional[str] = None,
    ):
        super().__init__()
        # TODO check if this is a good idea
        if isinstance(exit_condition, dict):
            exit_condition = load_pipeline_artefact_by_definition(exit_condition)

        # TODO: check if this is a good idea
        if len(pipeline_steps) > 0 and isinstance(pipeline_steps[0], dict):
            pipeline_steps = load_pipeline_steps_by_definition(pipeline_steps)

        self.pipeline_steps = pipeline_steps or []
        self.exit_condition = exit_condition

        if display_name:
            self.display_name = display_name


ExtensionType = TypeVar("ExtensionType", bound="ExtensionInterface")


class ExtensionInterface(BaseModel, Generic[ExtensionType]):
    @computed_field
    @property
    def name(self) -> str:
        return serialize_module_name(self)

    class GraphAccessMixin(ABC):
        pass


class GraphUpdateHook(BaseModel):
    """
    A hook that is executed before or after the graph is updated. This can be used to modify the graph before or after an update.
    """

    @abstractmethod
    def execute(
        self, graph: BaseGraphInterface, updates: List[TestResultInterface]
    ) -> Optional[List[TestResultInterface]]:
        """
        Execute the hook.
        :param graph: current state of the graph.
        :param updates: the planned/taken actions
        :return: A list of updates that should be applied to the graph if it is a pre hook. If it is a post hook, it should return None.
        """
        pass



import inspect
import logging
import platform
from abc import ABC
from copy import deepcopy
import time
from typing import Optional, List, Dict, Callable, Union, Any, Generator

import torch.multiprocessing as mp

from causy.data_loader import AbstractDataLoader
from causy.edge_types import DirectedEdge
from causy.graph import GraphManager, Graph, GraphError
from causy.graph_utils import unpack_run
from causy.interfaces import (
    PipelineStepInterface,
    LogicStepInterface,
    BaseGraphInterface,
    GraphModelInterface,
    GraphUpdateHook,
)
from causy.models import TestResultAction, Algorithm, ActionHistoryStep
from causy.variables import (
    resolve_variables_to_algorithm_for_pipeline_steps,
    resolve_variables,
    VariableType,
    VariableTypes,
)

logger = logging.getLogger(__name__)


class AbstractGraphModel(GraphModelInterface, ABC):
    """
    The graph model is the main class of causy. It is responsible for creating a graph from data and executing the pipeline_steps.

    The graph model is responsible for the following tasks:
    - Create a graph from data (create_graph_from_data)
    - Execute the pipeline_steps (execute_pipeline_steps)
    - Take actions on the graph (execute_pipeline_step & _take_action which is called by execute_pipeline_step)

    It also initializes and takes care of the multiprocessing pool.

    """

    algorithm: Algorithm
    pipeline_steps: List[PipelineStepInterface]
    graph: BaseGraphInterface
    pool: mp.Pool

    _graph_class: type[BaseGraphInterface]

    def __init__(
        self,
        graph=None,
        algorithm: Algorithm = None,
    ):
        self.graph = graph
        self.algorithm = algorithm

        if algorithm.pipeline_steps is not None:
            self.pipeline_steps = algorithm.pipeline_steps or []

        if self.__multiprocessing_required(self.pipeline_steps):
            self.pool = self.__initialize_pool()
        else:
            self.pool = None

    def __initialize_pool(self) -> mp.Pool:
        """
        Initialize the multiprocessing pool
        :return: the multiprocessing pool
        """
        # we need to set the start method to spawn because the default fork method does not work well with torch on linux
        # see https://pytorch.org/docs/stable/notes/multiprocessing.html
        if platform.system() == "Linux":
            try:
                mp.set_start_method("spawn")
            except RuntimeError:
                logger.warning(
                    "Could not set multiprocessing start method to spawn. Using default method. This might cause issues on Linux."
                )
        return mp.Pool(mp.cpu_count() * 2)

    def __multiprocessing_required(self, pipeline_steps):
        """
        Check if multiprocessing is required
        :param pipeline_steps: the pipeline steps
        :return: True if multiprocessing is required
        """
        for step in pipeline_steps:
            if hasattr(step, "parallel") and step.parallel:
                return True
        return False

    def __del__(self):
        if self.pool is not None:
            self.pool.close()
            self.pool.join()

    def __create_graph_from_dict(self, data: Dict[str, List[float]]):
        """
        Create a graph from a dictionary
        :param data: the dictionary
        :return: the graph
        """
        graph = GraphManager(graph_class=self._graph_class)
        for key, values in sorted(data.items()):
            graph.add_node(key, values, id_=key)
        return graph

    def __create_graph_from_list(self, data: List[Dict[str, float]]):
        """
        Create a graph from a list of dictionaries
        :param data:
        :return:
        """
        # initialize nodes
        keys = data[0].keys()
        nodes: Dict[str, List[float]] = {}

        for key in sorted(keys):
            nodes[key] = []

        # load nodes into node dict
        for row in data:
            for key in keys:
                nodes[key].append(row[key])

        graph = GraphManager(graph_class=self._graph_class)
        for key in keys:
            graph.add_node(key, nodes[key], id_=key)

        return graph

    def _create_from_data_loader(self, data_loader: AbstractDataLoader):
        """
        Create a graph from a data loader
        :param data_loader: the data loader
        :return: the graph
        """
        nodes: Dict[str, List[float]] = {}
        keys = None

        # load nodes into node dict
        for row in data_loader.load():
            if isinstance(row, dict) and "_dict" in row:
                # edge case for when data is in a dict of lists
                return self.__create_graph_from_dict(row["_dict"])

            if keys is None:
                keys = row.keys()
                for key in sorted(keys):
                    nodes[key] = []

            for key in keys:
                nodes[key].append(row[key])

        graph = GraphManager(graph_class=self._graph_class)
        for key in keys:
            graph.add_node(key, nodes[key], id_=key)

        return graph

    def create_graph_from_data(
        self,
        data: Union[List[Dict[str, float]], Dict[str, List[float]], AbstractDataLoader],
    ):
        """
        Create a graph from data
        :param data: is a list of dictionaries or a dictionary with lists
        :return:
        """

        if isinstance(data, AbstractDataLoader):
            graph = self._create_from_data_loader(data)
        elif isinstance(data, dict):
            graph = self.__create_graph_from_dict(data)
        else:
            graph = self.__create_graph_from_list(data)

        self.graph = graph
        return graph

    def create_all_possible_edges(self):
        """
        Create all possible edges on a graph
        TODO: replace me with the skeleton builders
        :return:
        """
        for u in self.graph.nodes.values():
            for v in self.graph.nodes.values():
                if u == v:
                    continue
                self.graph.add_edge(u, v, {})

    def execute_pipeline_steps(self) -> List[ActionHistoryStep]:
        """
        Execute all pipeline_steps
        :return: the steps taken during the step execution
        """
        all(self.execute_pipeline_step_with_progress())
        return self.graph.action_history

    def execute_pipeline_step_with_progress(self) -> Generator:
        started = time.time()
        for filter in self.pipeline_steps:
            logger.info(f"Executing pipeline step {filter.__class__.__name__}")
            if isinstance(filter, LogicStepInterface):
                actions_taken = filter.execute(self.graph.graph, self)
                self.graph.graph.action_history.append(actions_taken)
                continue
            yield {
                "step": filter.__class__.__name__,
                "previous_duration": time.time() - started,
            }
            started = time.time()
            actions_taken = self.execute_pipeline_step(filter)
            self.graph.graph.action_history.append(
                ActionHistoryStep(
                    name=filter.name,
                    actions=actions_taken,
                    duration=time.time() - started,
                )
            )

            self.graph.purge_soft_deleted_edges()

    def _format_yield(self, test_fn, graph, generator):
        """
        Format the yield for the parallel processing
        :param test_fn: the pipeline_step test function
        :param graph: the graph
        :param generator: the generator object which generates the combinations
        :return: yields the test function with its inputs
        """
        for i in generator:
            yield [test_fn, [*i], graph]

    def _execute_pre_graph_update_hooks(
        self,
        hooks: List[GraphUpdateHook],
        graph: BaseGraphInterface,
        results: List[TestResultAction],
    ) -> List[TestResultAction]:
        """
        Execute the graph update hooks
        :param graph: the graph
        :param hooks: the hooks
        :param results: the results
        :return:
        """
        if not hooks:
            return results
        for hook in hooks:
            results = hook.execute(graph, results)
        return results

    def _execute_post_graph_update_hooks(
        self,
        hooks: List[GraphUpdateHook],
        graph: BaseGraphInterface,
        results: List[TestResultAction],
    ) -> None:
        """
        Execute the graph update hooks
        :param graph: the graph
        :param hooks: the hooks
        :param results: the results
        :return:
        """
        if not hooks:
            return
        for hook in hooks:
            hook.execute(graph, results)

    def _execute_hook(self, hook: str, i: TestResultAction):
        """
        Execute a single hook
        :param hook: the name of the hook
        :param i: the test result action
        :return:
        """

        try:
            getattr(self.graph.get_edge(i.u, i.v).edge_type, hook)(self.graph, i)
        except GraphError as e:
            logger.info(f"Graph Error in hook ({str(hook)}): {e}")
        except Exception as e:
            logger.error(f"Error in hook ({str(hook)}): {e}")

    def _take_action(self, results, dry_run=False):
        """
        Take the actions returned by the test

        In causy changes on the graph are not executed directly. Instead, the test returns an action which should be executed on the graph.
        This is done to make it possible to execute the tests in parallel as well as to decide proactively at which point in the decisions taken by the pipeline step should be executed.
        Actions are returned by the test and are executed on the graph. The actions are stored in the action history to make it possible to revert the actions or use them in a later step.

        :param results:
        :return:
        """
        all_actions_taken = []
        for result_items in results:
            if result_items is None:
                continue
            if not isinstance(result_items, list):
                result_items = [result_items]
            result_items = self._execute_pre_graph_update_hooks(
                self.algorithm.pre_graph_update_hooks, self.graph, result_items
            )
            actions_taken = []
            for i in result_items:
                if i.u is not None and i.v is not None:
                    logger.debug(f"Action: {i.action} on {i.u.name} and {i.v.name}")

                if dry_run:
                    actions_taken.append(i)
                    continue

                # execute the action returned by the test
                if i.action == TestResultAction.REMOVE_EDGE_UNDIRECTED:
                    if not self.graph.undirected_edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to remove undirected edge {i.u.name} <-> {i.v.name}. But it does not exist."
                        )
                        continue
                    self._execute_hook("pre_edge_remove_hook", i)
                    self.graph.remove_edge(i.u, i.v, soft_delete=True)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self.graph.add_edge_history(i.v, i.u, i)
                    self._execute_hook("post_edge_remove_hook", i)
                elif i.action == TestResultAction.UPDATE_EDGE:
                    if not self.graph.edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update edge {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue
                    self._execute_hook("pre_edge_update_hook", i)
                    self.graph.update_edge(i.u, i.v, metadata=i.data)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self.graph.add_edge_history(i.v, i.u, i)
                    self._execute_hook("post_edge_update_hook", i)
                elif i.action == TestResultAction.UPDATE_EDGE_DIRECTED:
                    if not self.graph.directed_edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update directed edge {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue

                    self._execute_hook("pre_edge_update_hook", i)
                    self.graph.update_directed_edge(i.u, i.v, i.data)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self._execute_hook("post_edge_update_hook", i)
                elif i.action == TestResultAction.DO_NOTHING:
                    continue
                elif i.action == TestResultAction.REMOVE_EDGE_DIRECTED:
                    if not self.graph.directed_edge_exists(
                        i.u, i.v
                    ) and not self.graph.edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to remove directed edge {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue
                    self._execute_hook("pre_edge_remove_hook", i)
                    self.graph.remove_directed_edge(i.u, i.v, soft_delete=True)
                    # TODO: move this to pre/post update hooks
                    if self.graph.edge_exists(
                        i.v, i.u
                    ):  # if the edge is undirected, make it directed
                        self.graph.update_directed_edge(
                            i.v, i.u, edge_type=DirectedEdge()
                        )
                    self.graph.add_edge_history(i.u, i.v, i)
                    self._execute_hook("post_edge_remove_hook", i)
                elif i.action == TestResultAction.UPDATE_EDGE_TYPE:
                    if not self.graph.edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update edge type {i.u.name} <-> {i.v.name}. But it does not exist."
                        )
                        continue
                    self._execute_hook("pre_edge_type_change_hook", i)
                    self.graph.update_edge(i.u, i.v, edge_type=i.edge_type)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self.graph.add_edge_history(i.v, i.u, i)
                    self._execute_hook("post_edge_type_change_hook", i)
                elif i.action == TestResultAction.UPDATE_EDGE_TYPE_DIRECTED:
                    if not self.graph.directed_edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update edge type {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue
                    self._execute_hook("pre_edge_type_change_hook", i)
                    self.graph.update_directed_edge(i.u, i.v, edge_type=i.edge_type)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self._execute_hook("post_edge_type_change_hook", i)

                elif i.action == TestResultAction.RESTORE_EDGE:
                    if self.graph.edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to restore edge {i.u.name} <-> {i.v.name}. But it does exist."
                        )
                        continue

                    self.graph.restore_edge(i.u, i.v)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self.graph.add_edge_history(i.v, i.u, i)

                elif i.action == TestResultAction.RESTORE_EDGE_DIRECTED:
                    if self.graph.directed_edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to restore edge {i.u.name} <-> {i.v.name}. But it does exist."
                        )
                        continue

                    self.graph.restore_directed_edge(i.u, i.v)
                    self.graph.add_edge_history(i.u, i.v, i)
                # add the action to the actions history
                actions_taken.append(i)
            all_actions_taken.extend(actions_taken)
            self._execute_post_graph_update_hooks(
                self.algorithm.post_graph_update_hooks, self.graph, actions_taken
            )
        return all_actions_taken

    def execute_pipeline_step(
        self, test_fn: PipelineStepInterface, apply_to_graph=True
    ):
        """
        Execute a single pipeline_step on the graph. either in parallel or in a single process depending on the test_fn.parallel flag
        :param apply_to_graph:  if the action should be applied to the graph
        :param test_fn: the test function
        :param threshold: the threshold
        :return:
        """
        actions_taken = []
        # initialize the worker pool (we currently use all available cores * 2)

        # run all combinations in parallel except if the number of combinations is smaller then the chunk size
        # because then we would create more overhead then we would definetly gain from parallel processing
        if test_fn.parallel:
            if self.pool is None:
                logger.warning(
                    "Parallel processing is enabled but no pool is initialized. Initializing pool."
                )
                self.pool = self.__initialize_pool()

            for result in self.pool.imap_unordered(
                unpack_run,
                self._format_yield(
                    test_fn,
                    self.graph.graph,
                    test_fn.generator.generate(self.graph.graph, self),
                ),
                chunksize=test_fn.chunk_size_parallel_processing,
            ):
                if not isinstance(result, list):
                    result = [result]
                actions_taken.extend(
                    self._take_action(result, dry_run=not apply_to_graph)
                )
        else:
            if test_fn.generator.chunked:
                for chunk in test_fn.generator.generate(self.graph.graph, self):
                    iterator = [
                        unpack_run(i)
                        for i in [[test_fn, [*c], self.graph.graph] for c in chunk]
                    ]
                    actions_taken.extend(
                        self._take_action(iterator, dry_run=not apply_to_graph)
                    )
            else:
                # this is the only mode which supports unapplied actions to be passed to the next pipeline step (for now)
                # which are sometimes needed for e.g. conflict resolution
                iterator = [
                    i
                    for i in [
                        [test_fn, [*i], self.graph.graph]
                        for i in test_fn.generator.generate(self.graph.graph, self)
                    ]
                ]

                local_results = []
                for i in iterator:
                    rn_fn = i[0]
                    if hasattr(rn_fn, "needs_unapplied_actions"):
                        if rn_fn.needs_unapplied_actions:
                            i.append(local_results)
                    local_results.append(unpack_run(i))

                actions_taken.extend(
                    self._take_action(local_results, dry_run=not apply_to_graph)
                )

        return actions_taken


def graph_model_factory(
    algorithm: Algorithm = None,
    variables: Dict[str, VariableTypes] = None,
) -> type[AbstractGraphModel]:
    """
    Create a graph model based on a List of pipeline_steps
    :param algorithm: the algorithm which should be used to create the graph model
    :return: the graph model
    """
    original_algorithm = deepcopy(algorithm)
    if variables is None and algorithm.variables is not None:
        variables = resolve_variables(algorithm.variables, {})
    elif variables is None:
        variables = {}

    if len(variables) > 0:
        algorithm.pipeline_steps = resolve_variables_to_algorithm_for_pipeline_steps(
            algorithm.pipeline_steps, variables
        )

    bases = tuple()
    if algorithm.edge_types is not None:
        for edge_type in algorithm.edge_types:
            if not inspect.isabstract(edge_type.GraphAccessMixin):
                bases += (edge_type.GraphAccessMixin,)

    if algorithm.extensions is not None:
        for extension in algorithm.extensions:
            if not inspect.isabstract(extension.GraphAccessMixin):
                bases += (extension.GraphAccessMixin,)

    Graph.__bases__ = tuple(set(Graph.__bases__ + bases))

    class GraphModel(AbstractGraphModel):
        # store the original algorithm for later use like ejecting it without the resolved variables
        _original_algorithm = original_algorithm

        _graph_class = Graph

        def __init__(self):
            super().__init__(algorithm=algorithm)

    return GraphModel



import logging
import os
from datetime import datetime
from importlib.metadata import version
from typing import Dict, Any, List, Optional

import fastapi
import typer
import uvicorn

from causy.models import (
    AlgorithmReference,
    Algorithm,
    AlgorithmReferenceType,
)
from causy.serialization import load_algorithm_by_reference
from causy.ui.models import ExtendedResult, ExtendedExperiment, ExperimentVersion
from causy.workspaces.cli import (
    _load_latest_experiment_result,
    _load_experiment_versions,
    _load_experiment_result,
)
from causy.workspaces.models import Workspace
from fastapi import APIRouter, HTTPException
from starlette.staticfiles import StaticFiles

logger = logging.getLogger(__name__)
API_ROUTES = APIRouter()
MODEL: Optional[ExtendedResult] = None
WORKSPACE: Optional[Workspace] = None


@API_ROUTES.get("/status", response_model=Dict[str, Any])
async def get_status():
    """Get the current status of the API."""
    return {
        "status": "ok",
        "model_loaded": MODEL is not None,
        "workspace_loaded": WORKSPACE is not None,
        "mode": "workspace" if WORKSPACE else "model",
        "causy_version": version("causy"),
    }


@API_ROUTES.get("/model", response_model=ExtendedResult)
async def get_model():
    """Get the current model."""
    if not MODEL:
        raise HTTPException(404, "No model loaded")
    return MODEL


@API_ROUTES.get("/workspace", response_model=Workspace)
async def get_workspace():
    if not WORKSPACE:
        raise HTTPException(404, "No workspace loaded")
    return WORKSPACE


@API_ROUTES.get("/experiments/{experiment_name}/latest", response_model=ExtendedResult)
async def get_latest_experiment(experiment_name: str):
    """Get the current experiment."""
    if not WORKSPACE:
        raise HTTPException(404, "No workspace loaded")

    if experiment_name not in WORKSPACE.experiments:
        raise HTTPException(404, "Experiment not found")

    try:
        experiment = _load_latest_experiment_result(WORKSPACE, experiment_name)
    except Exception as e:
        raise HTTPException(400, str(e))

    version = _load_experiment_versions(WORKSPACE, experiment_name)[0]

    experiment["algorithm"] = AlgorithmReference(**experiment["algorithm"])
    experiment["version"] = version
    experiment = ExtendedResult(**experiment)

    return experiment


@API_ROUTES.get(
    "/experiments/{experiment_name}/{version_number}",
    response_model=ExtendedResult,
)
async def get_experiment(experiment_name: str, version_number: int):
    """Get the current experiment."""
    if not WORKSPACE:
        raise HTTPException(404, "No workspace loaded")

    if experiment_name not in WORKSPACE.experiments:
        raise HTTPException(404, "Experiment not found")

    try:
        experiment = _load_experiment_result(WORKSPACE, experiment_name, version_number)
    except Exception as e:
        raise HTTPException(400, str(e))

    experiment["algorithm"] = AlgorithmReference(**experiment["algorithm"])
    experiment["version"] = version_number
    experiment = ExtendedResult(**experiment)

    return experiment


@API_ROUTES.get("/experiments", response_model=List[ExtendedExperiment])
async def get_experiments():
    """Get the current experiment."""
    if not WORKSPACE:
        raise HTTPException(404, "No workspace loaded")

    experiments = []
    for experiment_name, experiment in WORKSPACE.experiments.items():
        extended_experiment = ExtendedExperiment(**experiment.model_dump())
        versions = []
        for experiment_version in _load_experiment_versions(WORKSPACE, experiment_name):
            versions.append(
                ExperimentVersion(
                    version=experiment_version,
                    name=datetime.fromtimestamp(experiment_version).isoformat(),
                )
            )
        extended_experiment.versions = versions
        extended_experiment.name = experiment_name
        experiments.append(extended_experiment)

    return experiments


@API_ROUTES.get("/algorithm/{reference_type}/{reference}", response_model=Algorithm)
async def get_algorithm(reference_type: str, reference: str):
    """Get the current algorithm."""
    if reference.startswith("/") or ".." in reference:
        raise HTTPException(400, "Invalid reference")

    if reference_type not in AlgorithmReferenceType.__members__.values():
        raise HTTPException(400, "Invalid reference type")

    try:
        algorithm = load_algorithm_by_reference(reference_type, reference)
        return algorithm
    except Exception as e:
        raise HTTPException(400, str(e))


def _create_ui_app(with_static=True):
    """Get the server."""
    app = fastapi.FastAPI(
        title="causy-api",
        version=version("causy"),
        description="causys internal api to serve data from the result files",
    )
    app.include_router(API_ROUTES, prefix="/api/v1", tags=["api"])
    if with_static:
        app.mount(
            "",
            StaticFiles(
                directory=os.path.join(os.path.dirname(__file__), "..", "static"),
                html=True,
            ),
            name="static",
        )
    return app


def _set_model(result: Dict[str, Any]):
    """Set the model."""
    global MODEL
    # for testing
    if result is None:
        MODEL = None
        return

    result["algorithm"] = AlgorithmReference(**result["algorithm"])
    MODEL = ExtendedResult(**result)


def _set_workspace(workspace: Workspace):
    """Set the workspace."""
    global WORKSPACE
    WORKSPACE = workspace


def is_port_in_use(host: str, port: int) -> bool:
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def server(result: Dict[str, Any] = None, workspace: Workspace = None):
    """Create the FastAPI server."""
    app = _create_ui_app()

    if result:
        _set_model(result)

    if workspace:
        _set_workspace(workspace)

    if not MODEL and not WORKSPACE:
        raise ValueError("No model or workspace provided")

    host = os.getenv("HOST", "localhost")
    is_port_from_env = os.getenv("PORT")
    if is_port_from_env:
        port = int(is_port_from_env)
    else:
        port = int(os.getenv("PORT", "8000"))
        while is_port_in_use(host, port):
            port += 1
            if port > 65535:
                raise ValueError("No free port available")

    cors_enabled = os.getenv("CORS_ENABLED", "false").lower() == "true"

    # cors e.g. for development of separate frontend
    if cors_enabled:
        logger.warning(typer.style("🌐 CORS enabled", fg=typer.colors.YELLOW))
        from fastapi.middleware.cors import CORSMiddleware

        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    server_config = uvicorn.Config(app, host=host, port=port, log_level="error")
    server = uvicorn.Server(server_config)
    return server_config, server



import copy
import datetime
import importlib
import json
from json import JSONEncoder
from typing import Dict, Any, List
import os

import torch
import yaml
from pydantic import TypeAdapter


from causy.edge_types import EDGE_TYPES
from causy.graph_utils import load_pipeline_steps_by_definition
from causy.models import AlgorithmReferenceType, Result, AlgorithmReference
from causy.variables import deserialize_variable
from causy.causal_discovery import AVAILABLE_ALGORITHMS


def serialize_algorithm(model, algorithm_name: str = None):
    """Serialize the model into a dictionary."""

    if algorithm_name:
        model.algorithm.name = algorithm_name
    return model.algorithm.model_dump()


def load_algorithm_from_specification(algorithm_dict: Dict[str, Any]):
    """Load the model from a dictionary."""
    algorithm_dict["pipeline_steps"] = load_pipeline_steps_by_definition(
        algorithm_dict["pipeline_steps"]
    )
    if "extensions" in algorithm_dict and algorithm_dict["extensions"] is not None:
        algorithm_dict["extensions"] = load_pipeline_steps_by_definition(
            algorithm_dict["extensions"]
        )
    else:
        algorithm_dict["extensions"] = []
    if "variables" not in algorithm_dict or algorithm_dict["variables"] is None:
        algorithm_dict["variables"] = []

    algorithm_dict["variables"] = [
        deserialize_variable(variable) for variable in algorithm_dict["variables"]
    ]
    from causy.models import Algorithm

    return TypeAdapter(Algorithm).validate_python(algorithm_dict)


def load_algorithm_by_reference(reference_type: str, algorithm: str):
    # TODO: test me
    if reference_type == AlgorithmReferenceType.FILE:
        # validate if the reference points only in the same directory or subdirectory
        # to avoid security issues
        absolute_path = os.path.realpath(algorithm)
        common_prefix = os.path.commonprefix([os.getcwd(), absolute_path])
        if common_prefix != os.getcwd():
            raise ValueError("Invalid reference")

        with open(absolute_path, "r") as file:
            # load the algorithm from the file
            # try first json
            try:
                algorithm = json.loads(file.read())
                return load_algorithm_from_specification(algorithm)
            except json.JSONDecodeError:
                pass
            file.seek(0)
            # then try yaml
            try:
                data = yaml.load(file.read(), Loader=yaml.FullLoader)
                return load_algorithm_from_specification(data)
            except yaml.YAMLError:
                pass
            raise ValueError("Invalid file format")

    elif reference_type == AlgorithmReferenceType.NAME:
        return copy.deepcopy(AVAILABLE_ALGORITHMS[algorithm]()._original_algorithm)
    elif reference_type == AlgorithmReferenceType.PYTHON_MODULE:
        module_, ref_ = algorithm.rsplit(".", 1)
        module = importlib.import_module(module_)
        st_function = getattr(module, ref_)
        if not st_function:
            raise ValueError(f"Algorithm {algorithm} not found")
        return st_function()._original_algorithm


class CausyJSONEncoder(JSONEncoder):
    def default(self, obj):
        if type(obj) is torch.Tensor:
            return None
        elif type(obj) is datetime.datetime:
            return obj.isoformat()
        return super().default(obj)


def load_json(pipeline_file: str):
    with open(pipeline_file, "r") as file:
        pipeline = json.loads(file.read())
    return pipeline


def deserialize_result(result: Dict[str, Any], klass=Result):
    """Deserialize the result."""

    result["algorithm"] = AlgorithmReference(**result["algorithm"])
    for i, edge in enumerate(result["edges"]):
        result["edges"][i]["edge_type"] = EDGE_TYPES[edge["edge_type"]["name"]](
            **edge["edge_type"]
        )

    return TypeAdapter(klass).validate_python(result)



import json
import logging
from collections import OrderedDict
from datetime import datetime
from typing import List, Dict

import pydantic_yaml
import questionary
import typer
import os
import sys

import yaml
from markdown.extensions.toc import slugify
from pydantic_yaml import to_yaml_str
from jinja2 import (
    Environment,
    select_autoescape,
    ChoiceLoader,
    FileSystemLoader,
    PackageLoader,
)
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from causy.graph_model import graph_model_factory
from causy.graph_utils import hash_dictionary
from causy.models import (
    AlgorithmReference,
    AlgorithmReferenceType,
    Result,
)
from causy.serialization import (
    load_algorithm_by_reference,
    CausyJSONEncoder,
    deserialize_result,
)
from causy.variables import validate_variable_values, resolve_variables
from causy.workspaces.models import Workspace, Experiment
from causy.data_loader import DataLoaderReference, load_data_loader

workspace_app = typer.Typer()

pipeline_app = typer.Typer()
experiment_app = typer.Typer()
dataloader_app = typer.Typer()
logger = logging.getLogger(__name__)

NO_COLOR = os.environ.get("NO_COLOR", False)

WORKSPACE_FILE_NAME = "workspace.yml"

JINJA_ENV = Environment(
    loader=ChoiceLoader(
        [
            PackageLoader("causy", "workspaces/templates"),
            FileSystemLoader("./templates"),
        ]
    ),
    autoescape=select_autoescape(),
)


class WorkspaceNotFoundError(Exception):
    pass


def show_error(message: str):
    if NO_COLOR:
        typer.echo(f"❌ {message}", err=True)
    else:
        typer.echo(typer.style(f"❌ {message}", fg=typer.colors.RED), err=True)


def show_success(message: str):
    typer.echo(f"✅ {message}")


def write_to_workspace(workspace: Workspace):
    workspace_path = os.path.join(os.getcwd(), WORKSPACE_FILE_NAME)
    ordered_val = OrderedDict(json.loads(workspace.model_dump_json()))
    yaml.add_representer(
        OrderedDict,
        lambda dumper, data: dumper.represent_mapping(
            "tag:yaml.org,2002:map", data.items()
        ),
    )
    output = yaml.dump(ordered_val)
    with open(workspace_path, "w") as f:
        f.write(output)


def _current_workspace(fail_if_none: bool = True) -> Workspace:
    """
    Return the current workspace.
    :param fail_if_none: if True, raise an exception if no workspace is found
    :return: the workspace
    """

    workspace_data = None
    workspace_path = os.path.join(os.getcwd(), WORKSPACE_FILE_NAME)
    # add the current directory to the path to allow for imports
    sys.path.append(os.getcwd())
    if os.path.exists(workspace_path):
        with open(workspace_path, "r") as f:
            workspace_data = f.read()

    if fail_if_none and workspace_data is None:
        raise WorkspaceNotFoundError("No workspace found in the current directory")

    workspace = None

    if workspace_data is not None:
        workspace = pydantic_yaml.parse_yaml_raw_as(Workspace, workspace_data)

    return workspace


def _create_pipeline(workspace: Workspace = None) -> Workspace:
    pipeline_creation = questionary.select(
        "Do you want to use an existing pipeline or create a new one?",
        choices=[
            questionary.Choice(
                "Use an existing causy pipeline (preconfigured).", "PRECONFIGURED"
            ),
            questionary.Choice(
                "Eject existing pipeline (allows you to change pipeline configs).",
                "EJECT",
            ),
            questionary.Choice(
                "Create a pipeline skeleton (as a python module).", "SKELETON"
            ),
        ],
    ).ask()

    if pipeline_creation == "PRECONFIGURED":
        from causy.causal_discovery.constraint.algorithms import AVAILABLE_ALGORITHMS

        pipeline_reference = questionary.select(
            "Which pipeline do you want to use?", choices=AVAILABLE_ALGORITHMS.keys()
        ).ask()
        pipeline_reference = AVAILABLE_ALGORITHMS[pipeline_reference]
        # make pipeline reference as string
        pipeline = AlgorithmReference(
            reference=pipeline_reference().algorithm.name,
            type=AlgorithmReferenceType.NAME,
        )

        pipeline_name = questionary.text("Enter the name of the pipeline").ask()
        pipeline_name = slugify(pipeline_name, "_")

        workspace.pipelines[pipeline_name] = pipeline
    elif pipeline_creation == "EJECT":
        from causy.causal_discovery.constraint.algorithms import AVAILABLE_ALGORITHMS

        pipeline_skeleton = questionary.select(
            "Which pipeline do you want to use?", choices=AVAILABLE_ALGORITHMS.keys()
        ).ask()
        pipeline_reference = AVAILABLE_ALGORITHMS[pipeline_skeleton]
        pipeline_name = questionary.text("Enter the name of the pipeline").ask()
        pipeline_slug = slugify(pipeline_name, "_")
        with open(f"{pipeline_slug}.yml", "w") as f:
            f.write(to_yaml_str(pipeline_reference()._original_algorithm))

        pipeline = AlgorithmReference(
            reference=f"{pipeline_slug}.yml", type=AlgorithmReferenceType.FILE
        )

        workspace.pipelines[pipeline_name] = pipeline
    elif pipeline_creation == "SKELETON":
        pipeline_name = questionary.text("Enter the name of the pipeline").ask()
        pipeline_slug = slugify(pipeline_name, "_")
        JINJA_ENV.get_template("pipeline.py.tpl").stream(
            pipeline_name=pipeline_name
        ).dump(f"{pipeline_slug}.py")
        pipeline = AlgorithmReference(
            reference=f"{pipeline_slug}.PIPELINE",
            type=AlgorithmReferenceType.PYTHON_MODULE,
        )
        workspace.pipelines[pipeline_slug] = pipeline

    typer.echo(f'Pipeline "{pipeline_name}" created.')

    return workspace


def _create_experiment(workspace: Workspace) -> Workspace:
    experiment_name = questionary.text("Enter the name of the experiment").ask()
    experiment_pipeline = questionary.select(
        "Select the pipeline for the experiment", choices=workspace.pipelines.keys()
    ).ask()
    experiment_data_loader = questionary.select(
        "Select the data loader for the experiment",
        choices=workspace.dataloaders.keys(),
    ).ask()

    experiment_slug = slugify(experiment_name, "_")

    # extract and prefill the variables
    variables = {}
    pipeline = load_algorithm_by_reference(
        workspace.pipelines[experiment_pipeline].type,
        workspace.pipelines[experiment_pipeline].reference,
    )
    if len(pipeline.variables) > 0:
        variables = resolve_variables(pipeline.variables, {})

    workspace.experiments[experiment_slug] = Experiment(
        **{
            "pipeline": experiment_pipeline,
            "dataloader": experiment_data_loader,
            "variables": variables,
        }
    )

    typer.echo(f'Experiment "{experiment_name}" created.')

    return workspace


def _create_data_loader(workspace: Workspace) -> Workspace:
    data_loader_type = questionary.select(
        "Do you want to use an existing pipeline or create a new one?",
        choices=[
            questionary.Choice("Load a JSON File.", "json"),
            questionary.Choice("Load a JSONL File.", "jsonl"),
            questionary.Choice("Load data dynamically (via Python Script).", "dynamic"),
        ],
    ).ask()

    data_loader_name = questionary.text("Enter the name of the data loader").ask()

    if data_loader_type in ["json", "jsonl"]:
        data_loader_path = questionary.path(
            "Choose the file or enter the file name:",
        ).ask()
        data_loader_slug = slugify(data_loader_name, "_")
        workspace.dataloaders[data_loader_slug] = DataLoaderReference(
            **{
                "type": data_loader_type,
                "reference": data_loader_path,
            }
        )
    elif data_loader_type == "dynamic":
        data_loader_slug = slugify(data_loader_name, "_")
        JINJA_ENV.get_template("dataloader.py.tpl").stream(
            data_loader_name=data_loader_name
        ).dump(f"{data_loader_slug}.py")
        workspace.dataloaders[data_loader_slug] = DataLoaderReference(
            **{
                "type": data_loader_type,
                "reference": f"{data_loader_slug}.DataLoader",
            }
        )

    typer.echo(f'Data loader "{data_loader_name}" created.')

    return workspace


def _execute_experiment(workspace: Workspace, experiment: Experiment) -> Result:
    """
    Execute an experiment. This function will load the pipeline and the data loader and execute the pipeline.
    :param workspace:
    :param experiment:
    :return:
    """
    typer.echo(f"Loading Pipeline: {experiment.pipeline}")
    pipeline = load_algorithm_by_reference(
        workspace.pipelines[experiment.pipeline].type,
        workspace.pipelines[experiment.pipeline].reference,
    )

    validate_variable_values(pipeline, experiment.variables)
    variables = resolve_variables(pipeline.variables, experiment.variables)
    typer.echo(f"Using variables: {variables}")

    typer.echo(f"Loading Data: {experiment.dataloader}")
    data_loader = load_data_loader(workspace.dataloaders[experiment.dataloader])
    model = graph_model_factory(pipeline, experiment.variables)()
    model.create_graph_from_data(data_loader)
    model.create_all_possible_edges()
    task_count = len(model.pipeline_steps)
    current = 0
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        prev_task = None
        prev_task_data = None
        for task in model.execute_pipeline_step_with_progress():
            current += 1
            if prev_task is not None:
                progress.update(
                    prev_task,
                    completed=True,
                    current=1,
                    description=f"✅ {prev_task_data['step']} ({round(task['previous_duration'])}s)",
                )
            prev_task = progress.add_task(description=task["step"], total=1)
            prev_task_data = task

    return Result(
        algorithm=workspace.pipelines[experiment.pipeline],
        action_history=model.graph.graph.action_history,
        edges=model.graph.retrieve_edges(),
        nodes=model.graph.nodes,
        variables=variables,
        data_loader_hash=data_loader.hash(),
        algorithm_hash=pipeline.hash(),
        variables_hash=hash_dictionary(variables),
    )


def _load_latest_experiment_result(
    workspace: Workspace, experiment_name: str
) -> Experiment:
    versions = _load_experiment_versions(workspace, experiment_name)

    if experiment_name not in workspace.experiments:
        raise ValueError(f"Experiment {experiment_name} not found in the workspace")

    if len(versions) == 0:
        raise ValueError(f"Experiment {experiment_name} not found in the file system")

    with open(f"{experiment_name}_{versions[0]}.json", "r") as f:
        experiment = json.load(f)

    return experiment


def _load_experiment_result(
    workspace: Workspace, experiment_name: str, version_number: int
) -> Dict[str, any]:
    if experiment_name not in workspace.experiments:
        raise ValueError(f"Experiment {experiment_name} not found in the workspace")

    if version_number not in _load_experiment_versions(workspace, experiment_name):
        raise ValueError(
            f"Version {version_number} not found for experiment {experiment_name}"
        )

    with open(f"{experiment_name}_{version_number}.json", "r") as f:
        experiment = json.load(f)

    return experiment


def _load_experiment_versions(workspace: Workspace, experiment_name: str) -> List[int]:
    versions = []
    for file in os.listdir():
        # check for files if they have the right prefix followed by a unix timestamp (int) and the file extension, e.g. experiment_123456789.json.
        # Extract the unix timestamp
        if file.startswith(f"{experiment_name}_") and file.endswith(".json"):
            segments = file.split("_")
            timestamp = int(segments[-1].split(".")[0])
            name = "_".join(segments[:-1])
            if name != experiment_name:
                # an experiment with a different name
                continue
            versions.append(timestamp)
    return sorted(versions, reverse=True)


def _save_experiment_result(workspace: Workspace, experiment_name: str, result: Result):
    timestamp = int(datetime.timestamp(result.created_at))
    with open(f"{experiment_name}_{timestamp}.json", "w") as f:
        f.write(json.dumps(result.model_dump(), cls=CausyJSONEncoder, indent=4))


@pipeline_app.command(name="add")
def create_pipeline():
    """Create a new pipeline in the current workspace."""
    workspace = _current_workspace()
    workspace = _create_pipeline(workspace)

    write_to_workspace(workspace)


@pipeline_app.command(name="rm")
def remove_pipeline(pipeline_name: str):
    """Remove a pipeline from the current workspace."""
    workspace = _current_workspace()

    if pipeline_name not in workspace.pipelines:
        show_error(f"Pipeline {pipeline_name} not found in the workspace.")
        return

    # check if the pipeline is still in use
    for experiment_name, experiment in workspace.experiments.items():
        if experiment.pipeline == pipeline_name:
            show_error(
                f"Pipeline {pipeline_name} is still in use by experiment {experiment_name}. Cannot remove."
            )
            return

    del workspace.pipelines[pipeline_name]

    write_to_workspace(workspace)

    show_success(f"Pipeline {pipeline_name} removed from the workspace.")


def _experiment_needs_reexecution(workspace: Workspace, experiment_name: str) -> bool:
    """
    Check if an experiment needs to be re-executed.
    :param workspace:
    :param experiment_name:
    :return:
    """
    if experiment_name not in workspace.experiments:
        raise ValueError(f"Experiment {experiment_name} not found in the workspace")

    versions = _load_experiment_versions(workspace, experiment_name)

    if len(versions) == 0:
        logger.info(f"Experiment {experiment_name} not found in the file system.")
        return True

    latest_experiment = _load_latest_experiment_result(workspace, experiment_name)
    experiment = workspace.experiments[experiment_name]
    latest_experiment = deserialize_result(latest_experiment)
    if (
        latest_experiment.algorithm_hash is None
        or latest_experiment.data_loader_hash is None
    ):
        logger.info(f"Experiment {experiment_name} has no hashes.")
        return True

    pipeline = load_algorithm_by_reference(
        workspace.pipelines[experiment.pipeline].type,
        workspace.pipelines[experiment.pipeline].reference,
    )

    validate_variable_values(pipeline, experiment.variables)
    variables = resolve_variables(pipeline.variables, experiment.variables)

    if latest_experiment.variables_hash != hash_dictionary(variables):
        logger.info(f"Experiment {experiment_name} has different variables.")
        return True

    model = graph_model_factory(pipeline, variables)()
    if latest_experiment.algorithm_hash != model.algorithm.hash():
        logger.info(f"Experiment {experiment_name} has a different pipeline.")
        return True

    data_loder = load_data_loader(workspace.dataloaders[experiment.dataloader])
    if latest_experiment.data_loader_hash != data_loder.hash():
        logger.info(
            f"Experiment {experiment_name} has a different data loader/dataset."
        )
        return True

    return False


def _clear_experiment(experiment_name: str, workspace: Workspace):
    versions = _load_experiment_versions(workspace, experiment_name)
    versions_removed = 0
    for version in versions:
        try:
            os.remove(f"{experiment_name}_{version}.json")
            versions_removed += 1
        except FileNotFoundError:
            pass
    return versions_removed


@experiment_app.command(name="add")
def create_experiment():
    """Create a new experiment in the current workspace."""
    workspace = _current_workspace()
    workspace = _create_experiment(workspace)

    write_to_workspace(workspace)


@experiment_app.command(name="rm")
def remove_experiment(experiment_name: str):
    """Remove an experiment from the current workspace."""
    workspace = _current_workspace()

    if experiment_name not in workspace.experiments:
        show_error(f"Experiment {experiment_name} not found in the workspace.")
        return

    versions_removed = _clear_experiment(experiment_name, workspace)

    del workspace.experiments[experiment_name]

    write_to_workspace(workspace)

    show_success(
        f"Experiment {experiment_name} removed from the workspace. Removed {versions_removed} versions."
    )


@experiment_app.command(name="clear")
def clear_experiment(experiment_name: str):
    """Clear all versions of an experiment."""
    workspace = _current_workspace()

    if experiment_name not in workspace.experiments:
        show_error(f"Experiment {experiment_name} not found in the workspace.")
        return

    versions_removed = _clear_experiment(experiment_name, workspace)

    write_to_workspace(workspace)

    show_success(
        f"Experiment {experiment_name} cleared. Removed {versions_removed} versions."
    )


@experiment_app.command(name="update-variable")
def update_experiment_variable(
    experiment_name: str, variable_name: str, variable_value: str
):
    """Update a variable in an experiment."""
    workspace = _current_workspace()

    if experiment_name not in workspace.experiments:
        show_error(f"Experiment {experiment_name} not found in the workspace.")
        return

    experiment = workspace.experiments[experiment_name]

    pipeline = load_algorithm_by_reference(
        workspace.pipelines[experiment.pipeline].type,
        workspace.pipelines[experiment.pipeline].reference,
    )

    current_variable = None
    for existing_variable in pipeline.variables:
        if variable_name == existing_variable.name:
            current_variable = existing_variable
            break
    else:
        show_error(f"Variable {variable_name} not found in the experiment.")
        return

    # try to cast the variable value to the correct type
    try:
        variable_value = current_variable._PYTHON_TYPE(variable_value)
    except ValueError:
        show_error(
            f'Variable {variable_name} should be {current_variable.type}. But got "{variable_value}" which is not a valid value.'
        )
        return

    # check if the variable is a valid value
    if not validate_variable_values(pipeline, {variable_name: variable_value}):
        show_error(f"Variable {variable_name} is not a valid value.")
        return

    experiment.variables[variable_name] = variable_value

    write_to_workspace(workspace)

    show_success(f"Variable {variable_name} updated in experiment {experiment_name}.")


@dataloader_app.command(name="add")
def create_data_loader():
    """Create a new data loader in the current workspace."""
    workspace = _current_workspace()
    workspace = _create_data_loader(workspace)

    write_to_workspace(workspace)


@dataloader_app.command(name="rm")
def remove_data_loader(data_loader_name: str):
    """Remove a data loader from the current workspace."""
    workspace = _current_workspace()

    if data_loader_name not in workspace.dataloaders:
        show_error(f"Data loader {data_loader_name} not found in the workspace.")
        return

    # check if the data loader is still in use
    for experiment_name, experiment in workspace.experiments.items():
        if experiment.dataloader == data_loader_name:
            show_error(
                f"Data loader {data_loader_name} is still in use by experiment {experiment_name}. Cannot remove."
            )
            return

    del workspace.dataloaders[data_loader_name]

    write_to_workspace(workspace)

    show_success(f"Data loader {data_loader_name} removed from the workspace.")


@workspace_app.command()
def info():
    """Show general information about the workspace."""
    workspace = _current_workspace()
    typer.echo(f"Workspace: {workspace.name}")
    typer.echo(f"Author: {workspace.author}")
    typer.echo(f"Pipelines: {workspace.pipelines}")
    typer.echo(f"Data loaders: {workspace.dataloaders}")
    typer.echo(f"Experiments: {workspace.experiments}")


@workspace_app.command()
def init():
    """
    Initialize a new workspace in the current directory.
    """
    workspace_path = os.path.join(os.getcwd(), WORKSPACE_FILE_NAME)
    sys.path.append(os.getcwd())

    if os.path.exists(workspace_path):
        typer.confirm(
            "Workspace already exists. Do you want to overwrite it?", abort=True
        )

    workspace = Workspace(
        **{
            "name": "",
            "author": "",
            "dataloaders": {},
            "pipelines": {},
            "experiments": {},
        }
    )

    current_folder_name = os.path.basename(os.getcwd())

    workspace.name = typer.prompt("Name", default=current_folder_name, type=str)
    workspace.author = typer.prompt(
        "Author", default=os.environ.get("USER", os.environ.get("USERNAME")), type=str
    )

    configure_pipeline = typer.confirm(
        "Do you want to configure a pipeline?", default=False
    )

    workspace.pipelines = {}
    if configure_pipeline:
        workspace = _create_pipeline(workspace)

    configure_data_loader = typer.confirm(
        "Do you want to configure a data loader?", default=False
    )

    workspace.dataloaders = {}
    if configure_data_loader:
        data_loader_type = questionary.select(
            "Do you want to use an existing pipeline or create a new one?",
            choices=[
                questionary.Choice("Load a JSON File.", "json"),
                questionary.Choice("Load a JSONL File.", "jsonl"),
                questionary.Choice(
                    "Load data dynamically (via Python Script).", "dynamic"
                ),
            ],
        ).ask()

        if data_loader_type in ["json", "jsonl"]:
            data_loader_path = questionary.path(
                "Choose the file or enter the file name:",
            ).ask()
            data_loader_name = questionary.text(
                "Enter the name of the data loader"
            ).ask()
            data_loader_slug = slugify(data_loader_name, "_")
            workspace.dataloaders[data_loader_slug] = DataLoaderReference(
                **{
                    "type": data_loader_type,
                    "reference": data_loader_path,
                }
            )
        elif data_loader_type == "dynamic":
            data_loader_name = questionary.text(
                "Enter the name of the data loader"
            ).ask()
            data_loader_slug = slugify(data_loader_name, "_")
            JINJA_ENV.get_template("dataloader.py.tpl").stream(
                data_loader_name=data_loader_name
            ).dump(f"{data_loader_slug}.py")
            workspace.dataloaders[data_loader_slug] = DataLoaderReference(
                **{
                    "type": data_loader_type,
                    "reference": f"{data_loader_slug}.DataLoader",
                }
            )
    workspace.experiments = {}

    if len(workspace.pipelines) > 0 and len(workspace.dataloaders) > 0:
        configure_experiment = typer.confirm(
            "Do you want to configure an experiment?", default=False
        )

        if configure_experiment:
            workspace = _create_experiment(workspace)

    with open(workspace_path, "w") as f:
        f.write(pydantic_yaml.to_yaml_str(workspace))

    typer.echo(f"Workspace created in {workspace_path}")


@workspace_app.command()
def execute(experiment_name: str = None, force_reexecution: bool = False):
    """
    Execute an experiment or all experiments in the workspace.
    """
    workspace = _current_workspace()
    if experiment_name is None:
        # execute all experiments
        for experiment_name, experiment in workspace.experiments.items():
            try:
                needs_reexecution = _experiment_needs_reexecution(
                    workspace, experiment_name
                )
            except ValueError as e:
                show_error(str(e))
                needs_reexecution = True

            if needs_reexecution is False and force_reexecution is False:
                typer.echo(f"Skipping experiment: {experiment_name}. (no changes)")
                continue
            typer.echo(f"Executing experiment: {experiment_name}")
            result = _execute_experiment(workspace, experiment)
            _save_experiment_result(workspace, experiment_name, result)
    else:
        if experiment_name not in workspace.experiments:
            typer.echo(f"Experiment {experiment_name} not found in the workspace.")
            return
        experiment = workspace.experiments[experiment_name]
        typer.echo(f"Executing experiment: {experiment_name}")
        result = _execute_experiment(workspace, experiment)

        _save_experiment_result(workspace, experiment_name, result)


@workspace_app.command()
def diff(experiment_names: List[str], only_differences: bool = False):
    """
    Show the differences between multiple experiment results.
    """
    workspace = _current_workspace()
    if len(experiment_names) < 2:
        show_error("Please provide at least two experiment names/versions.")
        return

    experiments_to_compare = []
    resolved_experiments = []

    # check if the experiment strings are experiments or experiment_versions and load the respective experiments/versions
    for experiment_name in experiment_names:
        if experiment_name not in workspace.experiments:
            potential_version = experiment_name.split("_")[-1]
            try:
                version = int(potential_version)
            except ValueError:
                show_error(f"Experiment {experiment_name} not found in the workspace")
                return
            experiment_name = "_".join(experiment_name.split("_")[:-1])
            if version not in _load_experiment_versions(workspace, experiment_name):
                show_error(
                    f"Version {version} not found for experiment {experiment_name}"
                )
                return
            experiment_result = deserialize_result(
                _load_experiment_result(workspace, experiment_name, version)
            )
            experiment = workspace.experiments[experiment_name]
            experiment_version = f"{experiment_name}_{version}"

        else:
            experiment = workspace.experiments[experiment_name]
            experiment_result = deserialize_result(
                _load_latest_experiment_result(workspace, experiment_name)
            )
            experiment_version = f"{experiment_name}_latest"

        experiments_to_compare.append(
            {
                "result": experiment_result,
                "experiment": experiment,
                "version": experiment_version,
            }
        )
        resolved_experiments.append(experiment_version)
    find_equivalents = {}

    # find the differences between all experiments and the differences for each of the other edges

    for experiment in experiments_to_compare:
        for edge in experiment["result"].edges:
            u, v = sorted([edge.u.name, edge.v.name])
            if u not in find_equivalents:
                find_equivalents[u] = {}
            if v not in find_equivalents[u]:
                find_equivalents[u][v] = {}

            if experiment["version"] in find_equivalents[u][v]:
                if find_equivalents[u][v][experiment["version"]] != edge:
                    typer.echo(
                        f"Experiment {experiment['experiment']} has an inconsistent edge {u} -> {v}"
                    )
            else:
                find_equivalents[u][v][experiment["version"]] = edge

    experiment_table = []

    for node_u, s in find_equivalents.items():
        for node_v, result in s.items():
            experiment_table_row = {exp: None for exp in resolved_experiments}
            for experiment, edge in result.items():
                experiment_table_row[experiment] = edge

            experiment_table.append(experiment_table_row)

    table = Table()
    table.add_column("Edge")

    for experiment in resolved_experiments:
        table.add_column(experiment, justify="center")

    for row in experiment_table:
        elements = [key for key in row.values()]
        first_element = None
        for e in elements:
            if e is not None:
                first_element = e
                break

        all_elements_same = all([e == first_element for e in elements])
        if only_differences and all_elements_same:
            continue
        table.add_row(
            *[f"{first_element.u.name} - {first_element.v.name}"]
            + [
                f"{row[experiment].edge_type.STR_REPRESENTATION}"
                if row[experiment]
                else ""
                for experiment in resolved_experiments
            ],
            style="green" if all_elements_same else "red",
        )

    console = Console()
    console.print(table)


workspace_app.add_typer(pipeline_app, name="pipeline", help="Manage pipelines")
workspace_app.add_typer(experiment_app, name="experiment", help="Manage experiments")
workspace_app.add_typer(dataloader_app, name="dataloader", help="Manage data loaders")



import abc
import copy
from dataclasses import dataclass
from types import SimpleNamespace

import torch
import random as python_random

from causy.graph import GraphManager
from typing import Dict, Callable, List, Optional, Union

import logging

logger = logging.getLogger(__name__)


def random_normal() -> float:
    """
    Returns a random number from a normal distribution
    :return: the random number as a float
    """
    return python_random.normalvariate(0, 1)


@dataclass
class NodeReference:
    """
    A reference to a node in the sample generator
    """

    node: str

    def __str__(self):
        return self.node


@dataclass
class TimeAwareNodeReference(NodeReference):
    """
    A reference to a node in the sample generator
    """

    point_in_time: int = 0

    def __str__(self):
        return f"{self.node} - t{self.point_in_time}"


@dataclass
class SampleEdge:
    """
    An edge in the sample generator that references a node and a lag
    """

    from_node: NodeReference
    to_node: NodeReference
    value: float = 0


class TimeTravelingError(Exception):
    """
    An error that is raised when a TimeProxy tries to access a value from the future
    """

    pass


class AbstractSampleGenerator(abc.ABC):
    def __init__(
        self,
        edges: List[Union[SampleEdge, SampleEdge]],
        random: Callable = random_normal,  # for setting that to a fixed value for testing use random = lambda: 0
    ):
        self._edges = edges
        self._variables = self._find_variables_in_edges()
        self.random_fn = random

    @abc.abstractmethod
    def generate(self, size: int) -> SimpleNamespace:
        """
        Generate data for a sample graph with a time dimension
        :param size: the number of time steps to generate
        :return: the generated data and the sample graph
        """
        pass

    @abc.abstractmethod
    def _generate_data(self, size: int) -> Dict[str, torch.Tensor]:
        """
        Generate data for a sample graph with a time dimension
        :param size: the number of time steps to generate
        :return: the generated data
        """
        pass

    def _find_variables_in_edges(self):
        """
        Find all variables in the edges of the sample generator. We need this to calculate the initial values for all existing variables.
        :return: a set of all variables submitted via the edges
        """
        variables = set()
        for edge in self._edges:
            variables.add(edge.from_node.node)
            variables.add(edge.to_node.node)
        return variables

    def _get_edges_for_node_to(self, node: str):
        """
        Get all edges that point to a specific node
        :param node: the node to get the edges for
        :return: a list of edges
        """
        return [edge for edge in self._edges if edge.to_node.node == node]


class IIDSampleGenerator(AbstractSampleGenerator):
    """
    A sample generator that generates data from i.i.d multivariate Gaussians.

    A variable can not depend on itself.

    Example:
    >>> sg = IIDSampleGenerator(
    >>>     edges=[
    >>>      SampleEdge(NodeReference("X"), NodeReference("Y"), 5),
    >>>      SampleEdge(NodeReference("Y"), NodeReference("Z"), 7),
    >>>      ],
    >>> )

    """

    def __init__(
        self,
        edges: List[Union[SampleEdge, SampleEdge]],
        random: Callable = random_normal,
    ):
        super().__init__(edges, random)
        self._variables = self.topologic_sort(
            sorted(copy.deepcopy(list(self._variables)))
        )

    def topologic_sort(self, nodes: List[str]):
        """
        Sorts the nodes topologically
        :param nodes: list of nodes
        :param edges: list of edges
        :return: a list of sorted nodes
        """
        sorted_nodes = []
        while nodes:
            for node in nodes:
                if set(
                    [edge.from_node.node for edge in self._get_edges_for_node_to(node)]
                ).issubset(set(sorted_nodes)):
                    sorted_nodes.append(node)
                    nodes.remove(node)
                    break
        return sorted_nodes

    def _generate_data(self, size):
        """
        Generate data for a sample graph with a time dimension
        :param size: the number of time steps to generate
        :return: the generated data
        """
        internal_repr = {}

        # Initialize the output dictionary by adding noise
        for k in self._variables:
            internal_repr[k] = torch.tensor(
                [self.random_fn() for _ in range(size)], dtype=torch.float64
            )

        # Iterate over the nodes and sort them by the number of ingoing edges
        ingoing_edges = {}

        # topological sort: sort the nodes in an order that the nodes that depend on other nodes come after the nodes they depend on
        for node_name in self._variables:
            # Get the edges that point to this node
            ingoing_edges[node_name] = self._get_edges_for_node_to(node_name)

        # Sort the node such that all nodes that appear in edges must have occured as keys before
        sorted_nodes = self._variables

        # Generate the data
        for to_node in sorted_nodes:
            for edge in ingoing_edges[to_node]:
                from_node = edge.from_node.node
                internal_repr[to_node] += edge.value * internal_repr[from_node]
        return internal_repr

    # TODO: adjust sample generator tests to this data shape
    def _generate_shaped_data(self, size):
        test_data = self._generate_data(size)
        data = {}
        for variable in test_data.keys():
            data[variable] = test_data[variable]

        test_data_shaped = []
        for i in range(size):
            entry = {}
            for key in data.keys():
                entry[key] = data[key][i]
            test_data_shaped.append(entry)
        return test_data_shaped

    def generate(self, size):
        """
        Generate data for a sample graph with a time dimension
        :param size: the number of time steps to generate
        :return: the generated data and the sample graph
        """
        output = self._generate_data(size)
        graph = GraphManager()
        for i in self._variables:
            graph.add_node(
                i,
                output[i],
                id_=i,
                metadata={"variable": i},
            )

        for edge in self._edges:
            graph.add_directed_edge(
                graph.nodes[edge.from_node.node],
                graph.nodes[edge.to_node.node],
                metadata={},
            )

        return output, graph


# TODO: Does not work for multiple lags properly yet and is numerically unstable for several cases (check why, fix it)
class TimeseriesSampleGenerator(AbstractSampleGenerator):
    """
    A sample generator that generates data for a sample graph with a time dimension.

    Edges are defined as SampleLaggedEdges, which define a directed edge from a source node to a target node with a
    ag. The lag is the number of time steps between the source and the target.

    A variable can depend on itself, but only on its past values (with a lag). This corresponds to autoregressive models.

    Example:
    >>> sg = TimeseriesSampleGenerator(
    >>>     edges=[
    >>>      SampleEdge(TimeAwareNodeReference("X", -1), TimeAwareNodeReference("X"), 0.9),
    >>>      SampleEdge(TimeAwareNodeReference("Y", -1), TimeAwareNodeReference("Y"), 0.9),
    >>>      SampleEdge(TimeAwareNodeReference("Z", -1), TimeAwareNodeReference("Z"), 0.9),
    >>>      SampleEdge(TimeAwareNodeReference("Z", -1), TimeAwareNodeReference("Y"), 5),
    >>>      SampleEdge(TimeAwareNodeReference("Y", -1), TimeAwareNodeReference("X"), 7),
    >>>      ],
    >>> )
    """

    def __init__(
        self,
        edges: List[Union[SampleEdge, SampleEdge]],
        random: Callable = random_normal,  # for setting that to a fixed value for testing use random = lambda: 0
    ):
        super().__init__(edges, random)
        self._longest_lag = max(
            [abs(edge.from_node.point_in_time) for edge in self._edges]
        )

    _initial_distribution_fn: Callable = lambda self, x: torch.normal(0, x)

    def _generate_data(self, size):
        """
        Generate data for a sample graph with a time dimension
        :param size: the number of time steps to generate
        :return: the generated data
        """
        internal_repr = {}

        initial_values = self._calculate_initial_values()
        # Initialize the output dictionary
        for k in self._variables:
            internal_repr[k] = [initial_values[k]]

        for t in range(1, size):
            for node_name in self._variables:
                # Get the edges that point to this node
                edges = self._get_edges_for_node_to(node_name)
                result = torch.tensor(0.0, dtype=torch.float64)
                for edge in edges:
                    if abs(edge.from_node.point_in_time) > t:
                        result += (
                            edge.value * initial_values[edge.from_node.node]
                        )  # TODO(sofia): map here to the proper point in time
                    else:
                        result += (
                            edge.value
                            * internal_repr[edge.from_node.node][
                                t + edge.from_node.point_in_time
                            ]
                        )

                result += self.random_fn()
                internal_repr[node_name].append(result)

        for k, v in internal_repr.items():
            internal_repr[k] = torch.stack(v)

        return internal_repr

    def generate(self, size):
        """
        Generate data for a sample graph with a time dimension
        :param size: the number of time steps to generate
        :return: the generated data and the sample graph
        """
        output = self._generate_data(size)
        graph = GraphManager()
        for i in self._variables:
            for t in range(size):
                graph.add_node(
                    f"{i} - t{t}",
                    [output[i][t]],
                    id_=f"{i}-t{t}",
                    metadata={"time": t, "variable": i},
                )

        for t in range(1, size):
            for edge in self._edges:
                if t - abs(edge.from_node.point_in_time) < 0:
                    logger.debug(
                        f"Cannot generate data for {edge.from_node.node} at t={t}, "
                        f"since it depends on {abs(edge.to_node.point_in_time)}-steps-ago value"
                    )
                else:
                    graph.add_directed_edge(
                        graph.nodes[
                            f"{edge.from_node.node}-t{t - abs(edge.from_node.point_in_time)}"
                        ],
                        graph.nodes[f"{edge.to_node.node}-t{t}"],
                        metadata={},
                    )

        return output, graph

    def custom_block_diagonal(self, matrices):
        # Get dimensions
        num_matrices = len(matrices)
        n = len(matrices[0])  # Assuming all matrices are of the same size

        # Compute total size of the block diagonal matrix
        total_rows = num_matrices * n
        total_cols = num_matrices * n

        # Create an empty tensor to hold the block diagonal matrix
        block_diag = torch.zeros(total_rows, total_cols)

        # Fill in the first rows with the input matrices
        for i, matrix in enumerate(matrices):
            block_diag[:n, i * n : (i + 1) * n] = torch.tensor(
                matrix, dtype=torch.float64
            )

        # Fill in the lower left off-diagonal with identity matrices
        row_start = n
        col_start = 0
        for i in range(1, num_matrices):
            block_diag[
                row_start : row_start + n, col_start : col_start + n
            ] = torch.eye(n)
            row_start += n
            col_start += n

        return block_diag

    def __generate_coefficient_matrix(self):
        """
        generate the coefficient matrix for the sample generator graph
        :return: the coefficient matrix
        """

        matrix: List[List[List[float]]] = [
            [[0 for _ in self._variables] for _ in self._variables]
            for _ in range(self._longest_lag)
        ]

        # map the initial values to numbers from 0 to n
        values_map = self.__matrix_position_mapping()

        for edge in self._edges:
            matrix[(edge.from_node.point_in_time * -1) - 1][
                values_map[edge.to_node.node]
            ][values_map[edge.from_node.node]] = edge.value

        # return me as torch tensor
        return self.custom_block_diagonal(matrix)

    def vectorize_identity_block(self, n):
        # Create an empty tensor
        matrix = torch.zeros(n, n)

        # Fill the upper left block with an identity matrix
        matrix[: n // self._longest_lag, : n // self._longest_lag] = torch.eye(
            n // self._longest_lag
        )

        # Flatten the matrix
        vectorized_matrix = matrix.view(-1)

        return vectorized_matrix

    def _calculate_initial_values(self):
        """
        Calculate the initial values for the sample generator graph using the covariance matrix of the noise terms.

        coefficient_matrix=[[a,0],[b,a]], i.e.
        coefficient_matrix[0][0] is the coefficient of X_t-1 in the equation for X_t, here a
        coefficient_matrix[0][1] is the coefficient of Y_t-1 in the equation for X_t (here: no edge, that means zero)
        coefficient_matrix[1][1] is the coefficient of Y_t-1 in the equation for Y_t, here a
        coefficient_matrix[1][0] is the coefficient of X_t-1 in the equation for Y_t, here b

        If the top left entry ([0][0]) in our coefficient matrix corresponds to X_{t-1} -> X_t, then the the top left
        entry in the covariance matrix is the variance of X_t, i.e. V(X_t).

        If the top right entry ([0][1]) in our coefficient matrix corresponds to X_{t-1} -> Y_t, then the the top left
        entry in the covariance matrix is the variance of X_t, i.e. V(X_t).
        :return: the initial values for the sample generator graph as a dictionary
        """
        coefficient_matrix = self.__generate_coefficient_matrix()
        n, _ = coefficient_matrix.shape

        kronecker_product = torch.kron(coefficient_matrix, coefficient_matrix)
        identity_matrix = torch.eye(n**2)
        cov_matrix_noise_terms_vectorized = self.vectorize_identity_block(n)
        inv_identity_minus_kronecker_product = torch.linalg.pinv(
            identity_matrix - kronecker_product
        )
        vectorized_covariance_matrix = torch.matmul(
            inv_identity_minus_kronecker_product,
            cov_matrix_noise_terms_vectorized,
        )
        vectorized_covariance_matrix = vectorized_covariance_matrix.reshape(n, n)

        initial_values: Dict[str, torch.Tensor] = {}
        values = torch.diagonal(vectorized_covariance_matrix, offset=0)
        for k, i in self.__matrix_position_mapping().items():
            initial_values[k] = self._initial_distribution_fn(torch.sqrt(values[i]))
        return initial_values

    def __matrix_position_mapping(self):
        """
        Map the variables to numbers from 0 to n. This is needed to calculate the initial values for the sample generator graph.
        :return:
        """
        values_map = {}
        for i, k in enumerate(self._variables):
            values_map[k] = i
        return values_map



import itertools
import math
from typing import Tuple, List, Optional, Generic
import logging

import torch
from scipy import stats

from causy.generators import AllCombinationsGenerator, PairsWithNeighboursGenerator
from causy.math_utils import get_t_and_critical_t
from causy.interfaces import (
    PipelineStepInterface,
    BaseGraphInterface,
    NodeInterface,
    AS_MANY_AS_FIELDS,
    GeneratorInterface,
    PipelineStepInterfaceType,
)
from causy.models import ComparisonSettings, TestResultAction, TestResult
from causy.variables import IntegerParameter, BoolParameter

logger = logging.getLogger(__name__)


class CorrelationCoefficientTest(
    PipelineStepInterface[PipelineStepInterfaceType], Generic[PipelineStepInterfaceType]
):
    generator: Optional[GeneratorInterface] = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing: IntegerParameter = 1
    parallel: BoolParameter = False

    def process(
        self, nodes: List[str], graph: BaseGraphInterface
    ) -> Optional[TestResult]:
        """
        Test if u and v are independent and delete edge in graph if they are.
        :param nodes: list of nodes
        :return: A TestResult with the action to take
        """
        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]

        # Use Fisher's Z test for independency of u and v
        sample_size = len(x.values)
        corr = graph.edge_value(x, y)["correlation"]

        # Fisher's Z transformation
        fisher_z = 0.5 * math.log((1 + corr) / (1 - corr))

        # Use absolute value of fisher_z for a two-tailed test
        z_value = abs(fisher_z) * math.sqrt(sample_size - 3)

        # Compute the p-value
        p_value = 2 * (1 - stats.norm.cdf(z_value))

        # If the p value is smaller than the threshold, the null hypothesis (independence) is rejected, otherwise we delete the edge
        if p_value > self.threshold:
            logger.debug(f"Nodes {x.name} and {y.name} are uncorrelated")
            logger.debug(f"P-value: {p_value}")
            return TestResult(
                u=x,
                v=y,
                action=TestResultAction.REMOVE_EDGE_UNDIRECTED,
                data={},
            )


class PartialCorrelationTest(
    PipelineStepInterface[PipelineStepInterfaceType], Generic[PipelineStepInterfaceType]
):
    generator: Optional[GeneratorInterface] = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=3, max=3)
    )
    chunk_size_parallel_processing: IntegerParameter = 1
    parallel: BoolParameter = False

    def process(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult]]:
        """
        Test if nodes u,v are independent given node z based on a partial correlation test.
        We use this test for all combinations of 3 nodes because it is faster than the extended test (which supports combinations of n nodes). We can
        use it to remove edges between nodes which are not independent given another node and so reduce the number of combinations for the extended test.
        :param nodes: the nodes to test
        :return: A TestResult with the action to take

        TODO: we are testing (C and E given B) and (E and C given B), we just need one of these, remove redundant tests.
        """
        results = []
        already_deleted_edges = set()
        for nodes in itertools.permutations(nodes):
            x: NodeInterface = graph.nodes[nodes[0]]
            y: NodeInterface = graph.nodes[nodes[1]]
            z: NodeInterface = graph.nodes[nodes[2]]

            # Avoid division by zero
            if x is None or y is None or z is None:
                return

            if not graph.edge_exists(x, y) or (y, x) in already_deleted_edges:
                continue

            try:
                cor_xy = graph.edge_value(x, y)["correlation"]
                cor_xz = graph.edge_value(x, z)["correlation"]
                cor_yz = graph.edge_value(y, z)["correlation"]
            except (KeyError, TypeError):
                return

            numerator = cor_xy - cor_xz * cor_yz
            denominator = ((1 - cor_xz**2) * (1 - cor_yz**2)) ** 0.5

            # Avoid division by zero
            if denominator == 0:
                return

            par_corr = numerator / denominator

            # make t test for independency of u and v given z
            sample_size = len(x.values)
            nb_of_control_vars = len(nodes) - 2

            # Fisher's Z transformation
            fisher_z = 0.5 * math.log((1 + par_corr) / (1 - par_corr))

            # Use absolute value of fisher_z for a two-tailed test
            z_value = abs(fisher_z) * math.sqrt(sample_size - nb_of_control_vars - 3)

            # Compute the p-value
            p_value = 2 * (1 - stats.norm.cdf(z_value))

            # If the p value is smaller than the threshold, the null hypothesis (independence) is rejected, otherwise we delete the edge
            if p_value > self.threshold:
                results.append(
                    TestResult(
                        u=x,
                        v=y,
                        action=TestResultAction.REMOVE_EDGE_UNDIRECTED,
                        data={"separatedBy": [z]},
                    )
                )
                already_deleted_edges.add((x, y))

        return results


class ExtendedPartialCorrelationTestMatrix(
    PipelineStepInterface[PipelineStepInterfaceType], Generic[PipelineStepInterfaceType]
):
    generator: Optional[GeneratorInterface] = PairsWithNeighboursGenerator(
        comparison_settings=ComparisonSettings(min=4, max=AS_MANY_AS_FIELDS),
        shuffle_combinations=False,
    )
    chunk_size_parallel_processing: IntegerParameter = 1000
    parallel: BoolParameter = False

    def process(
        self, nodes: List[str], graph: BaseGraphInterface
    ) -> Optional[TestResult]:
        """
        Test if nodes u,v are independent given Z (set of nodes) based on partial correlation using the inverted covariance matrix (precision matrix).
        https://en.wikipedia.org/wiki/Partial_correlation#Using_matrix_inversion
        We use this test for all combinations of more than 3 nodes because it is slower.
        If the covariance matrix is ill-conditioned, i.e., its condition number is high, the precision matrix is not reliable.
        In that case, we throw a warning.
        :param nodes: the nodes to test
        :return: A TestResult with the action to take
        """

        if not graph.edge_exists(graph.nodes[nodes[0]], graph.nodes[nodes[1]]):
            return

        other_neighbours = set(
            [
                k
                for k, value in graph.edges[nodes[0]].items()
                if graph.directed_edge_exists(k, nodes[0])
            ]
        )
        other_neighbours.remove(graph.nodes[nodes[1]].id)

        if not set(nodes[2:]).issubset(set([on for on in list(other_neighbours)])):
            return
        cov_matrix = torch.cov(
            torch.stack([graph.nodes[node].values for node in nodes])
        )
        # check if the covariance matrix is ill-conditioned
        if torch.det(cov_matrix) == 0:
            logger.warning(
                "The covariance matrix is ill-conditioned. The precision matrix is not reliable."
            )
            return

        inverse_cov_matrix = torch.inverse(cov_matrix)

        n = inverse_cov_matrix.size(0)
        diagonal = torch.diag(inverse_cov_matrix)
        diagonal_matrix = torch.zeros((n, n), dtype=torch.float64)
        for i in range(n):
            diagonal_matrix[i, i] = diagonal[i]

        helper = torch.mm(torch.sqrt(diagonal_matrix), inverse_cov_matrix)
        precision_matrix = torch.mm(helper, torch.sqrt(diagonal_matrix))

        par_corr = (-1 * precision_matrix[0][1]) / torch.sqrt(
            precision_matrix[0][0] * precision_matrix[1][1]
        )

        sample_size = len(graph.nodes[nodes[0]].values)
        nb_of_control_vars = len(nodes) - 2

        # Fisher's Z transformation
        fisher_z = 0.5 * math.log((1 + par_corr) / (1 - par_corr))

        # Use absolute value of fisher_z for a two-tailed test
        z_value = abs(fisher_z) * math.sqrt(sample_size - nb_of_control_vars - 3)

        # Compute the p-value
        p_value = 2 * (1 - stats.norm.cdf(z_value))

        if p_value > self.threshold:
            logger.debug(
                f"Nodes {graph.nodes[nodes[0]].name} and {graph.nodes[nodes[1]].name} are uncorrelated given nodes {','.join([graph.nodes[on].name for on in other_neighbours])}"
            )
            nodes_set = set([graph.nodes[n] for n in nodes])
            return TestResult(
                u=graph.nodes[nodes[0]],
                v=graph.nodes[nodes[1]],
                action=TestResultAction.REMOVE_EDGE_UNDIRECTED,
                data={
                    "separatedBy": list(
                        nodes_set - {graph.nodes[nodes[0]], graph.nodes[nodes[1]]}
                    )
                },
            )


def partial_correlation_regression(x, y, z):
    """
    Compute the partial correlation coefficient between x and y controlling for other variables in z using linear regression.

    Arguments:
    x, y : torch.Tensor : Variables for which the partial correlation is computed.
    z : torch.Tensor : Other variables used to control for in the partial correlation.

    Returns:
    partial_corr : torch.Tensor : Partial correlation coefficient between x and y.
    """
    # Define linear regression model
    model_x = torch.linalg.lstsq(z.T, x).solution
    model_y = torch.linalg.lstsq(z.T, y).solution

    residual_x = x - torch.matmul(model_x, z)
    residual_y = y - torch.matmul(model_y, z)

    # Compute correlation of residuals
    return torch.dot(residual_x, residual_y) / (
        torch.norm(residual_x) * torch.norm(residual_y)
    )


class ExtendedPartialCorrelationTestLinearRegression(
    PipelineStepInterface[PipelineStepInterfaceType], Generic[PipelineStepInterfaceType]
):
    generator: Optional[GeneratorInterface] = PairsWithNeighboursGenerator(
        comparison_settings=ComparisonSettings(min=4, max=AS_MANY_AS_FIELDS),
        shuffle_combinations=False,
    )
    chunk_size_parallel_processing: IntegerParameter = 1000
    parallel: BoolParameter = False

    def process(
        self, nodes: List[str], graph: BaseGraphInterface
    ) -> Optional[TestResult]:
        if not graph.edge_exists(graph.nodes[nodes[0]], graph.nodes[nodes[1]]):
            return

        other_neighbours = set(graph.edges[nodes[0]])
        other_neighbours.remove(graph.nodes[nodes[1]].id)

        partial_correlation = partial_correlation_regression(
            graph.nodes[nodes[0]].values,
            graph.nodes[nodes[1]].values,
            torch.stack([graph.nodes[node].values for node in nodes[2:]]),
        )

        sample_size = len(graph.nodes[nodes[0]].values)
        nb_of_control_vars = len(nodes) - 2

        t, critical_t = get_t_and_critical_t(
            sample_size,
            nb_of_control_vars,
            partial_correlation.item(),
            self.threshold,
        )

        if abs(t) < critical_t:
            logger.debug(
                f"Nodes {graph.nodes[nodes[0]].name} and {graph.nodes[nodes[1]].name} are uncorrelated given nodes {','.join([graph.nodes[on].name for on in other_neighbours])}"
            )
            nodes_set = set([graph.nodes[n] for n in nodes])
            return TestResult(
                u=graph.nodes[nodes[0]],
                v=graph.nodes[nodes[1]],
                action=TestResultAction.REMOVE_EDGE_UNDIRECTED,
                data={
                    "separatedBy": list(
                        nodes_set - {graph.nodes[nodes[0]], graph.nodes[nodes[1]]}
                    )
                },
            )



from causy.causal_effect_estimation.multivariate_regression import (
    ComputeDirectEffectsInDAGsMultivariateRegression,
)
from causy.common_pipeline_steps.exit_conditions import ExitOnNoActions
from causy.contrib.graph_ui import GraphUIExtension
from causy.edge_types import (
    DirectedEdge,
    UndirectedEdge,
    DirectedEdgeUIConfig,
    UndirectedEdgeUIConfig,
)
from causy.generators import PairsWithNeighboursGenerator, RandomSampleGenerator
from causy.graph_model import graph_model_factory
from causy.common_pipeline_steps.logic import Loop, ApplyActionsTogether
from causy.causal_discovery.constraint.independence_tests.common import (
    CorrelationCoefficientTest,
    PartialCorrelationTest,
    ExtendedPartialCorrelationTestMatrix,
)
from causy.common_pipeline_steps.calculation import (
    CalculatePearsonCorrelations,
)
from causy.interfaces import AS_MANY_AS_FIELDS
from causy.models import ComparisonSettings, Algorithm
from causy.causal_discovery.constraint.orientation_rules.pc import (
    ColliderTest,
    NonColliderTest,
    FurtherOrientTripleTest,
    OrientQuadrupleTest,
    FurtherOrientQuadrupleTest,
)
from causy.variables import (
    FloatVariable,
    VariableReference,
    IntegerVariable,
)

PC_DEFAULT_THRESHOLD = 0.005

PC_ORIENTATION_RULES = [
    ColliderTest(display_name="Collider Test"),
    Loop(
        pipeline_steps=[
            NonColliderTest(display_name="Non-Collider Test"),
            FurtherOrientTripleTest(display_name="Further Orient Triple Test"),
            OrientQuadrupleTest(display_name="Orient Quadruple Test"),
            FurtherOrientQuadrupleTest(display_name="Further Orient Quadruple Test"),
        ],
        display_name="Orientation Rules Loop",
        exit_condition=ExitOnNoActions(),
    ),
]

PC_GRAPH_UI_EXTENSION = GraphUIExtension(
    edges=[
        DirectedEdgeUIConfig(),
        UndirectedEdgeUIConfig(),
    ]
)

PC_EDGE_TYPES = [DirectedEdge(), UndirectedEdge()]

PC = graph_model_factory(
    Algorithm(
        pipeline_steps=[
            CalculatePearsonCorrelations(display_name="Calculate Pearson Correlations"),
            CorrelationCoefficientTest(
                threshold=VariableReference(name="threshold"),
                display_name="Correlation Coefficient Test",
            ),
            PartialCorrelationTest(
                threshold=VariableReference(name="threshold"),
                display_name="Partial Correlation Test",
            ),
            ExtendedPartialCorrelationTestMatrix(
                threshold=VariableReference(name="threshold"),
                display_name="Extended Partial Correlation Test Matrix",
            ),
            *PC_ORIENTATION_RULES,
        ],
        edge_types=PC_EDGE_TYPES,
        extensions=[PC_GRAPH_UI_EXTENSION],
        name="PC",
        variables=[FloatVariable(name="threshold", value=PC_DEFAULT_THRESHOLD)],
    )
)

PCStable = graph_model_factory(
    Algorithm(
        pipeline_steps=[
            CalculatePearsonCorrelations(),
            ApplyActionsTogether(
                pipeline_steps=[
                    CorrelationCoefficientTest(
                        threshold=VariableReference(name="threshold")
                    ),
                    PartialCorrelationTest(
                        threshold=VariableReference(name="threshold"),
                    ),
                    ExtendedPartialCorrelationTestMatrix(
                        threshold=VariableReference(name="threshold"),
                    ),
                ]
            ),
            *PC_ORIENTATION_RULES,
        ],
        edge_types=PC_EDGE_TYPES,
        extensions=[PC_GRAPH_UI_EXTENSION],
        name="PCStable",
        variables=[FloatVariable(name="threshold", value=PC_DEFAULT_THRESHOLD)],
    )
)


ParallelPC = graph_model_factory(
    Algorithm(
        pipeline_steps=[
            CalculatePearsonCorrelations(display_name="Calculate Pearson Correlations"),
            CorrelationCoefficientTest(
                threshold=VariableReference(name="threshold"),
                display_name="Correlation Coefficient Test",
            ),
            PartialCorrelationTest(
                threshold=VariableReference(name="threshold"),
                parallel=True,
                chunk_size_parallel_processing=50000,
                display_name="Partial Correlation Test",
            ),
            ExtendedPartialCorrelationTestMatrix(
                # run first a sampled version of the test so we can minimize the number of tests in the full version
                threshold=VariableReference(name="threshold"),
                display_name="Sampled Extended Partial Correlation Test Matrix",
                chunk_size_parallel_processing=5000,
                parallel=True,
                generator=RandomSampleGenerator(
                    generator=PairsWithNeighboursGenerator(
                        chunked=False,
                        shuffle_combinations=True,
                        comparison_settings=ComparisonSettings(
                            min=4, max=AS_MANY_AS_FIELDS
                        ),
                    ),
                    chunked=False,
                    every_nth=VariableReference(name="every_nth_sample"),
                ),
            ),
            ExtendedPartialCorrelationTestMatrix(
                threshold=VariableReference(name="threshold"),
                display_name="Extended Partial Correlation Test Matrix",
                chunk_size_parallel_processing=20000,
                parallel=True,
                generator=PairsWithNeighboursGenerator(
                    chunked=False,
                    shuffle_combinations=True,
                    comparison_settings=ComparisonSettings(
                        min=4, max=AS_MANY_AS_FIELDS
                    ),
                ),
            ),
            *PC_ORIENTATION_RULES,
        ],
        edge_types=PC_EDGE_TYPES,
        extensions=[PC_GRAPH_UI_EXTENSION],
        name="ParallelPC",
        variables=[
            FloatVariable(name="threshold", value=PC_DEFAULT_THRESHOLD),
            IntegerVariable(name="every_nth_sample", value=200),
        ],
    )
)



from typing import Optional, Dict

from pydantic import BaseModel

from causy.data_loader import DataLoaderReference
from causy.models import AlgorithmReference, Algorithm
from causy.variables import VariableType


class Experiment(BaseModel):
    """
    represents a single experiment
    :param name: name of the experiment
    :param pipeline: the name of the pipeline used
    """

    pipeline: str
    dataloader: str
    variables: Optional[Dict[str, VariableType]] = {}


class Workspace(BaseModel):
    name: str
    author: Optional[str]

    pipelines: Optional[Dict[str, Algorithm | AlgorithmReference]]
    dataloaders: Optional[Dict[str, DataLoaderReference]]
    experiments: Optional[Dict[str, Experiment]]



from typing import Tuple, Optional

import torch

from causy.generators import PairsWithEdgesInBetweenGenerator
from causy.interfaces import (
    BaseGraphInterface,
    PipelineStepInterface,
    GeneratorInterface,
)
from causy.models import TestResultAction, TestResult


class ComputeDirectEffectsInDAGsMultivariateRegression(PipelineStepInterface):
    generator: Optional[GeneratorInterface] = PairsWithEdgesInBetweenGenerator()

    chunk_size_parallel_processing: int = 1
    parallel: bool = False

    def process(self, nodes: Tuple[str], graph: BaseGraphInterface) -> TestResult:
        """
        Calculate the direct effect of each edge in the graph using multivariate regression.
        :param nodes: list of nodes
        :return: A TestResult with the action to take
        """
        effect_variable = graph.nodes[nodes[1]]
        cause_variable = graph.nodes[nodes[0]]

        # if there is an undirected edge, the effect is not identifiable
        if graph.undirected_edge_exists(cause_variable, effect_variable):
            edge_data = graph.edge_value(cause_variable, effect_variable)
            if "direct_effect" in edge_data:
                return
            edge_data["direct_effect"] = None
            return TestResult(
                u=graph.nodes[nodes[0]],
                v=graph.nodes[nodes[1]],
                action=TestResultAction.UPDATE_EDGE_DIRECTED,
                data=edge_data,
            )

        edge_data = graph.edge_value(cause_variable, effect_variable)

        all_parents_of_effect_variable = graph.parents_of_node(effect_variable)
        parents_of_effect_variable_without_cause_variable = [
            parent
            for parent in all_parents_of_effect_variable
            if parent != cause_variable
        ]

        effect_variable_parents_values = torch.stack(
            [cause_variable.values]
            + [
                parents_of_effect_variable_without_cause_variable[i].values
                for i in range(len(parents_of_effect_variable_without_cause_variable))
            ],
            dim=1,
        )
        # multivariate regression via tridiagonal reduction and SVD, see torch documentation
        reshaped_effect_variable_values = effect_variable.values.view(-1, 1)

        coefficients = torch.linalg.lstsq(
            effect_variable_parents_values,
            reshaped_effect_variable_values,
            driver="gelsd",
        ).solution

        edge_data["direct_effect"] = coefficients[0][0].item()

        return TestResult(
            u=graph.nodes[nodes[0]],
            v=graph.nodes[nodes[1]],
            action=TestResultAction.UPDATE_EDGE_DIRECTED,
            data=edge_data,
        )
